#!/usr/bin/env python
#
# render.py - Generate screenshots of overlays using OpenGL.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""The ``render`` module is a program which provides off-screen rendering
capability for scenes which can otherwise be displayed via *FSLeyes*.
"""


import os.path as op
import            sys
import            logging
import            textwrap

import numpy as np

import fsl.utils.idle                        as idle
import fsleyes_widgets.utils.layout          as fsllayout

import                                          fsleyes
import fsleyes.version                       as version
import fsleyes.overlay                       as fsloverlay
import fsleyes.colourmaps                    as fslcm
import fsleyes.parseargs                     as parseargs
import fsleyes.displaycontext                as displaycontext
import fsleyes.displaycontext.orthoopts      as orthoopts
import fsleyes.displaycontext.lightboxopts   as lightboxopts
import fsleyes.displaycontext.scene3dopts    as scene3dopts
import fsleyes.controls.colourbar            as cbar
import fsleyes.gl                            as fslgl
import fsleyes.gl.textures.imagetexture      as imagetexture
import fsleyes.gl.ortholabels                as ortholabels
import fsleyes.gl.offscreenslicecanvas       as slicecanvas
import fsleyes.gl.offscreenlightboxcanvas    as lightboxcanvas
import fsleyes.gl.offscreenscene3dcanvas     as scene3dcanvas


log = logging.getLogger(__name__)


def main(args=None):
    """Entry point for ``render``.

    Creates and renders an OpenGL scene, and saves it to a file, according
    to the specified command line arguments (which default to
    ``sys.argv[1:]``).
    """

    if args is None:
        args = sys.argv[1:]

    # Initialise FSLeyes and implement hacks.
    # This must come first as, amongst other
    # things, it sets the fsleyes.assetDir.
    fsleyes.initialise()

    # Initialise colour maps module
    fslcm.init()

    # Create a GL context
    fslgl.getGLContext(offscreen=True, createApp=True)

    # Now that GL inititalisation is over,
    # make sure that the idle loop executes
    # all tasks synchronously, instead of
    # trying to schedule them on the wx
    # event loop. And make sure image textures
    # don't use separate threads for data
    # processing.
    with idle.idleLoop.synchronous(), \
         imagetexture.ImageTexture.enableThreading(False):

        # Parse arguments, and
        # configure logging/debugging
        namespace = parseArgs(args)
        fsleyes.configLogging(namespace.verbose, namespace.noisy)

        # Initialise the fsleyes.gl modules
        fslgl.bootstrap(namespace.glversion)

        # Create a description of the scene
        overlayList, displayCtx, sceneOpts = makeDisplayContext(namespace)

        import matplotlib.image as mplimg

        # Render that scene, and save it to file
        bitmap, bg = render(namespace, overlayList, displayCtx, sceneOpts)

        if namespace.crop is not None:
            bitmap = autocrop(bitmap, bg, namespace.crop)
        mplimg.imsave(namespace.outfile, bitmap)


def parseArgs(argv):
    """Creates an argument parser which accepts options for off-screen
    rendering. Uses the :mod:`fsleyes.parseargs` module to peform the
    actual parsing.

    :returns: An ``argparse.Namespace`` object containing the parsed
              arguments.
    """

    mainParser = parseargs.ArgumentParser(
        add_help=False,
        formatter_class=parseargs.FSLeyesHelpFormatter)

    mainParser.add_argument('-of',
                            '--outfile',
                            help='Output image file name'),
    mainParser.add_argument('-c',
                            '--crop',
                            type=int,
                            metavar='BORDER',
                            help='Auto-crop image, leaving a '
                                 'border on each side')
    mainParser.add_argument('-sz',
                            '--size',
                            type=int, nargs=2,
                            metavar=('W', 'H'),
                            help='Size in pixels (width, height)',
                            default=(800, 600))

    name        = 'render'
    prolog      = 'FSLeyes render version {}\n'.format(version.__version__)
    optStr      = '-of outfile'
    description = textwrap.dedent("""\
        FSLeyes screenshot generator.

        Use the '--scene' option to choose between orthographic
        ('ortho'), lightbox ('lightbox'), or 3D ('3d') views.
        """)

    namespace = parseargs.parseArgs(
        mainParser,
        argv,
        name,
        prolog=prolog,
        desc=description,
        usageProlog=optStr,
        argOpts=['-of', '--outfile',
                 '-sz', '--size',
                 '-c',  '--crop'],
        shortHelpExtra=['--outfile', '--size', '--crop'])

    if namespace.outfile is None:
        log.error('outfile is required')
        mainParser.print_usage()
        sys.exit(1)

    namespace.outfile = op.abspath(namespace.outfile)

    if namespace.scene not in ('ortho', 'lightbox', '3d'):
        log.info('Unknown scene specified  ("{}") - defaulting '
                 'to ortho'.format(namespace.scene))
        namespace.scene = 'ortho'

    return namespace


def makeDisplayContext(namespace):
    """Creates :class:`.OverlayList`, :class:`.DisplayContext``, and
    :class:`.SceneOpts` instances which represent the scene to be rendered,
    as described by the arguments in the given ``namespace`` object.
    """

    # Create an overlay list and display context.
    # The DisplayContext, Display and DisplayOpts
    # classes are designed to be created in a
    # parent-child hierarchy. So we need to create
    # a 'dummy' master display context to make
    # things work properly.
    overlayList      = fsloverlay.OverlayList()
    masterDisplayCtx = displaycontext.DisplayContext(overlayList)
    childDisplayCtx  = displaycontext.DisplayContext(overlayList,
                                                     parent=masterDisplayCtx)

    # We have to artificially create a ref to the
    # master display context, otherwise it may get
    # gc'd arbitrarily. The parent reference in the
    # child creation above is ultimately stored as
    # a weakref, so we need to create a real one.
    childDisplayCtx.masterDisplayCtx = masterDisplayCtx

    # The handleOverlayArgs function uses the
    # fsleyes.overlay.loadOverlays function,
    # which will call these functions as it
    # goes through the list of overlay to be
    # loaded.
    def load(ovl):
        log.info('Loading overlay {} ...'.format(ovl))

    def error(ovl, error):
        log.error('Error loading overlay {}: {}'.format(ovl, error))
        raise error

    # Load the overlays specified on the command
    # line, and configure their display properties
    parseargs.applyMainArgs(   namespace,
                               overlayList,
                               masterDisplayCtx)
    parseargs.applyOverlayArgs(namespace,
                               overlayList,
                               masterDisplayCtx,
                               loadFunc=load,
                               errorFunc=error)

    # Create a SceneOpts instance describing
    # the scene to be rendered. The parseargs
    # module assumes that GL canvases have
    # already been created, so we use mock
    # objects to trick it. The options applied
    # to these mock objects are applied to the
    # real canvases later on, in the render
    # function below.
    if namespace.scene == 'ortho':
        sceneOpts = orthoopts.OrthoOpts(MockCanvasPanel(3))
    elif namespace.scene == 'lightbox':
        sceneOpts = lightboxopts.LightBoxOpts(MockCanvasPanel(1))
    elif namespace.scene == '3d':
        sceneOpts = scene3dopts. Scene3DOpts(MockCanvasPanel(1))

    # 3D views default to
    # world display space
    if namespace.scene == '3d':
        childDisplayCtx.displaySpace = 'world'

    parseargs.applySceneArgs(namespace,
                             overlayList,
                             childDisplayCtx,
                             sceneOpts)

    # Centre the location. The DisplayContext
    # will typically centre its location on
    # initialisation, but this may not work
    # if any overlay arguments change the bounds
    # of an overlay (e.g. mesh reference image)
    if namespace.worldLoc is None and namespace.voxelLoc is None:
        b = childDisplayCtx.bounds
        childDisplayCtx.location = [
            b.xlo + 0.5 * b.xlen,
            b.ylo + 0.5 * b.ylen,
            b.zlo + 0.5 * b.zlen]

    # This has to be applied after applySceneArgs,
    # in case the user used the '-std'/'-std1mm'
    # options.
    if namespace.selectedOverlay is not None:
        masterDisplayCtx.selectedOverlay = namespace.selectedOverlay

    if len(overlayList) == 0:
        raise RuntimeError('At least one overlay must be specified')

    return overlayList, childDisplayCtx, sceneOpts


def render(namespace, overlayList, displayCtx, sceneOpts):
    """Renders the scene, and returns a tuple containing the bitmap and the
    background colour.

    :arg namespace:   ``argparse.Namespace`` object containing command line
                      arguments.

    :arg overlayList: The :class:`.OverlayList` instance.
    :arg displayCtx:  The :class:`.DisplayContext` instance.
    :arg sceneOpts:   The :class:`.SceneOpts` instance.
    """

    # Calculate canvas and colour bar sizes
    # so that the entire scene will fit in
    # the width/height specified by the user
    width, height = namespace.size
    (width, height), (cbarWidth, cbarHeight) = \
        adjustSizeForColourBar(width,
                               height,
                               sceneOpts.showColourBar,
                               sceneOpts.colourBarLocation,
                               sceneOpts.labelSize)

    # Lightbox view -> only one canvas
    if namespace.scene == 'lightbox':
        c = createLightBoxCanvas(namespace,
                                 width,
                                 height,
                                 overlayList,
                                 displayCtx,
                                 sceneOpts)
        canvases = [c]

    # Ortho view -> up to three canvases
    elif namespace.scene == 'ortho':
        canvases = createOrthoCanvases(namespace,
                                       width,
                                       height,
                                       overlayList,
                                       displayCtx,
                                       sceneOpts)
        labelMgr = ortholabels.OrthoLabels(overlayList,
                                           displayCtx,
                                           sceneOpts,
                                           *canvases)
        labelMgr.refreshLabels()

    # 3D -> one 3D canvas
    elif namespace.scene == '3d':
        c = create3DCanvas(namespace,
                           width,
                           height,
                           overlayList,
                           displayCtx,
                           sceneOpts)

        canvases = [c]

    # Do we need to do a neuro/radio l/r flip?
    if namespace.scene in ('ortho', 'lightbox'):
        inRadio = displayCtx.displaySpaceIsRadiological()
        lrFlip  = displayCtx.radioOrientation != inRadio

        if lrFlip:
            for c in canvases:
                if c.opts.zax in (1, 2):
                    c.opts.invertX = True

    # fix orthographic projection if
    # showing an ortho grid layout.
    # Note that, if the user chose 'grid',
    # but also chose to hide one or more
    # canvases, the createOrthoCanvases
    # function will have adjusted the
    # value of sceneOpts.layout. So
    # if layout == grid, we definitely
    # have three canvases.
    #
    # The createOrthoCanvases also
    # re-orders the canvases, which
    # we're assuming knowledge of,
    # by indexing canvases[1].
    if namespace.scene == 'ortho' and sceneOpts.layout == 'grid':
        canvases[1].opts.invertX = True

    # Configure each of the canvases (with those
    # properties that are common to both ortho and
    # lightbox canvases) and render them one by one
    canvasBmps = []

    for i, c in enumerate(canvases):

        c.opts.pos = displayCtx.location

        # HACK If a SliceCanvas/LightBoxCanvas
        # is rendering the sceen to an off-screen
        # texture due to the low performance
        # setting, its internal viewport will not
        # be set until after all GLObjects have
        # been rendered. But some GLObjects (e.g.
        # GLLabel) need to know the current
        # viewport.
        #
        # This is very much an edge case, as who
        # would be using a low performance setting
        # for off-screen rendering?

        if namespace.scene in ('ortho', 'lightbox') and \
           namespace.performance is not None        and \
           int(namespace.performance) < 3:
            c._setViewport()

        c.draw()

        canvasBmps.append(c.getBitmap())

    # destroy the canvases
    for c in canvases:
        c.destroy()
    canvases = None

    # layout the bitmaps
    if namespace.scene in ('lightbox', '3d'):
        layout = fsllayout.Bitmap(canvasBmps[0])
    elif len(canvasBmps) > 0:
        layout = fsllayout.buildOrthoLayout(canvasBmps,
                                            None,
                                            sceneOpts.layout,
                                            False,
                                            0)
    else:
        layout = fsllayout.Space(width, height)

    # Render a colour bar if required
    if sceneOpts.showColourBar:
        cbarBmp = buildColourBarBitmap(overlayList,
                                       displayCtx,
                                       cbarWidth,
                                       cbarHeight,
                                       sceneOpts)
        if cbarBmp is not None:
            layout  = buildColourBarLayout(layout,
                                           cbarBmp,
                                           sceneOpts.colourBarLocation,
                                           sceneOpts.colourBarLabelSide)

    # Turn the layout tree into a bitmap image
    bgColour = [c * 255 for c in sceneOpts.bgColour]
    return fsllayout.layoutToBitmap(layout, bgColour), bgColour


def createLightBoxCanvas(namespace,
                         width,
                         height,
                         overlayList,
                         displayCtx,
                         sceneOpts):
    """Creates, configures, and returns an :class:`.OffScreenLightBoxCanvas`.

    :arg namespace:   ``argparse.Namespace`` object.
    :arg width:       Available width in pixels.
    :arg height:      Available height in pixels.
    :arg overlayList: The :class:`.OverlayList` instance.
    :arg displayCtx:  The :class:`.DisplayContext` instance.
    :arg sceneOpts:   The :class:`.SceneOpts` instance.
    """

    canvas = lightboxcanvas.OffScreenLightBoxCanvas(
        overlayList,
        displayCtx,
        zax=sceneOpts.zax,
        width=width,
        height=height)

    if sceneOpts.zrange == (0, 0):
        sceneOpts.zrange = displayCtx.bounds.getRange(sceneOpts.zax)

    opts                = canvas.opts
    opts.showCursor     = sceneOpts.showCursor
    opts.bgColour       = sceneOpts.bgColour
    opts.cursorColour   = sceneOpts.cursorColour
    opts.renderMode     = sceneOpts.renderMode
    opts.zax            = sceneOpts.zax
    opts.sliceSpacing   = sceneOpts.sliceSpacing
    opts.nrows          = sceneOpts.nrows
    opts.ncols          = sceneOpts.ncols
    opts.zrange         = sceneOpts.zrange
    opts.showGridLines  = sceneOpts.showGridLines
    opts.highlightSlice = sceneOpts.highlightSlice

    return canvas


def createOrthoCanvases(namespace,
                        width,
                        height,
                        overlayList,
                        displayCtx,
                        sceneOpts):
    """Creates, configures, and returns up to three
    :class:`.OffScreenSliceCanvas` instances, for rendering the scene.

    :arg namespace:   ``argparse.Namespace`` object.
    :arg width:       Available width in pixels.
    :arg height:      Available height in pixels.
    :arg overlayList: The :class:`.OverlayList` instance.
    :arg displayCtx:  The :class:`.DisplayContext` instance.
    :arg sceneOpts:   The :class:`.SceneOpts` instance.
    """

    canvases = []

    # Build a list containing the horizontal
    # and vertical axes for each canvas
    canvasAxes = []
    zooms      = []
    centres    = []
    if sceneOpts.showXCanvas:
        canvasAxes.append((1, 2))
        zooms     .append(sceneOpts.xzoom)
        centres   .append(sceneOpts.panel.getGLCanvases()[0].centre)
    if sceneOpts.showYCanvas:
        canvasAxes.append((0, 2))
        zooms     .append(sceneOpts.yzoom)
        centres   .append(sceneOpts.panel.getGLCanvases()[1].centre)
    if sceneOpts.showZCanvas:
        canvasAxes.append((0, 1))
        zooms     .append(sceneOpts.zzoom)
        centres   .append(sceneOpts.panel.getGLCanvases()[2].centre)

    # Grid layout only makes sense if
    # we're displaying 3 canvases
    if sceneOpts.layout == 'grid' and len(canvasAxes) <= 2:
        sceneOpts.layout = 'horizontal'

    if sceneOpts.layout == 'grid':
        canvasAxes = [canvasAxes[1], canvasAxes[0], canvasAxes[2]]
        centres    = [centres[   1], centres[   0], centres[   2]]
        zooms      = [zooms[     1], zooms[     0], zooms[     2]]

    # Calculate the size in pixels for each canvas
    sizes = calculateOrthoCanvasSizes(overlayList,
                                      displayCtx,
                                      width,
                                      height,
                                      canvasAxes,
                                      sceneOpts.layout)

    # Configure the properties on each canvas
    for ((width, height), (xax, yax), zoom, centre) in zip(sizes,
                                                           canvasAxes,
                                                           zooms,
                                                           centres):

        zax = 3 - xax - yax

        c = slicecanvas.OffScreenSliceCanvas(
            overlayList,
            displayCtx,
            zax=zax,
            width=int(width),
            height=int(height))

        opts              = c.opts
        opts.showCursor   = sceneOpts.showCursor
        opts.cursorColour = sceneOpts.cursorColour
        opts.cursorGap    = sceneOpts.cursorGap
        opts.bgColour     = sceneOpts.bgColour
        opts.renderMode   = sceneOpts.renderMode

        if zoom is not None:
            opts.zoom = zoom

        # Default to centering
        # on the cursor
        if centre is None:
            centre = [displayCtx.location[xax], displayCtx.location[yax]]

        c.centreDisplayAt(*centre)

        canvases.append(c)


    return canvases


def create3DCanvas(namespace,
                   width,
                   height,
                   overlayList,
                   displayCtx,
                   sceneOpts):
    """Creates, configures, and returns an :class:`.OffScreenScene3DCanvas`.

    :arg namespace:   ``argparse.Namespace`` object.
    :arg width:       Available width in pixels.
    :arg height:      Available height in pixels.
    :arg overlayList: The :class:`.OverlayList` instance.
    :arg displayCtx:  The :class:`.DisplayContext` instance.
    :arg sceneOpts:   The :class:`.SceneOpts` instance.
    """

    canvas = scene3dcanvas.OffScreenScene3DCanvas(
        overlayList,
        displayCtx,
        width=width,
        height=height)

    opts                 = canvas.opts
    opts.showCursor      = sceneOpts.showCursor
    opts.cursorColour    = sceneOpts.cursorColour
    opts.bgColour        = sceneOpts.bgColour
    opts.showLegend      = sceneOpts.showLegend
    opts.legendColour    = sceneOpts.fgColour
    opts.occlusion       = sceneOpts.occlusion
    opts.light           = sceneOpts.light
    opts.zoom            = sceneOpts.zoom
    opts.offset          = sceneOpts.offset
    opts.rotation        = sceneOpts.rotation

    if parseargs.wasSpecified(namespace, sceneOpts, 'lightPos'):
        opts.lightPos        = sceneOpts.lightPos
        canvas.resetLightPos = False
    else:
        canvas.defaultLightPos()

    return canvas


def buildColourBarBitmap(overlayList,
                         displayCtx,
                         width,
                         height,
                         sceneOpts):
    """If the currently selected overlay has a display range,
    creates and returns a bitmap containing a colour bar. Returns
    ``None`` otherwise.

    :arg overlayList:   The :class:`.OverlayList`.

    :arg displayCtx:    The :class:`.DisplayContext`.

    :arg width:         Colour bar width in pixels.

    :arg height:        Colour bar height in pixels.

    :arg sceneOpts:     :class:`.SceneOpts` instance containing display
                        settings.
    """

    overlay = displayCtx.getSelectedOverlay()
    display = displayCtx.getDisplay(overlay)
    opts    = display.opts

    cbarLocation = sceneOpts.colourBarLocation
    cbarSize     = sceneOpts.colourBarSize

    if   cbarLocation in ('top', 'bottom'): width  = width  * cbarSize / 100.0
    elif cbarLocation in ('left', 'right'): height = height * cbarSize / 100.0

    if not isinstance(opts, displaycontext.ColourMapOpts):
        return None

    if   cbarLocation in ('top', 'bottom'): orient = 'horizontal'
    elif cbarLocation in ('left', 'right'): orient = 'vertical'

    cb = cbar.ColourBar(overlayList, displayCtx)
    cb.orientation = orient
    cb.labelSide   = sceneOpts.colourBarLabelSide
    cb.bgColour    = sceneOpts.bgColour
    cb.textColour  = sceneOpts.fgColour
    cb.fontSize    = sceneOpts.labelSize

    cbarBmp = cb.colourBar(width, height)

    # The colourBarBitmap function returns a w*h*4
    # array, but the fsleyes_widgets.utils.layout.Bitmap
    # (see the next function) assumes a h*w*4 array
    cbarBmp = cbarBmp.transpose((1, 0, 2))

    return cbarBmp


def buildColourBarLayout(canvasLayout,
                         cbarBmp,
                         cbarLocation,
                         cbarLabelSide):
    """Given a layout object containing the rendered canvas bitmaps,
    creates a new layout which incorporates the given colour bar bitmap.

    :arg canvasLayout:  An object describing the canvas layout (see
                        :mod:`fsleyes_widgets.utils.layout`)

    :arg cbarBmp:       A bitmap containing a rendered colour bar.

    :arg cbarLocation:  Colour bar location (see :func:`buildColourBarBitmap`).

    :arg cbarLabelSide: Colour bar label side (see
                        :func:`buildColourBarBitmap`).
    """

    cbarBmp = fsllayout.Bitmap(cbarBmp)

    if   cbarLocation in ('top',    'left'):  items = [cbarBmp, canvasLayout]
    elif cbarLocation in ('bottom', 'right'): items = [canvasLayout, cbarBmp]

    if   cbarLocation in ('top', 'bottom'): return fsllayout.VBox(items)
    elif cbarLocation in ('left', 'right'): return fsllayout.HBox(items)


def adjustSizeForColourBar(width,
                           height,
                           showColourBar,
                           colourBarLocation,
                           fontSize):
    """Calculates the widths and heights of the image display space, and the
    colour bar if it is enabled.

    :arg width:             Desired width in pixels

    :arg height:            Desired height in pixels

    :arg showColourBar:     ``True`` if a colour bar is to be shown, ``False``
                            otherwise.

    :arg colourBarLocation: Colour bar location (see
                            :func:`buildColourBarBitmap`).

    :arg fontSize           Font size (points) used in colour bar labels.

    :returns:               Two tuples - the first tuple contains the
                            ``(width, height)`` of the available canvas space,
                            and the second contains the ``(width, height)`` of
                            the colour bar.
    """

    if showColourBar:

        cbarWidth = int(round(cbar.colourBarMinorAxisSize(fontSize)))
        if colourBarLocation in ('top', 'bottom'):
            height     = height - cbarWidth
            cbarHeight = cbarWidth
            cbarWidth  = width
        else:
            width      = width  - cbarWidth
            cbarHeight = height
    else:
        cbarWidth  = 0
        cbarHeight = 0

    return (width, height), (cbarWidth, cbarHeight)


def calculateOrthoCanvasSizes(overlayList,
                              displayCtx,
                              width,
                              height,
                              canvasAxes,
                              layout):
    """Calculates the sizes, in pixels, for each canvas to be displayed in an
    orthographic layout.

    :arg overlayList: The :class:`.OverlayList`.

    :arg displayCtx:  The :class:`.DisplayContext`.

    :arg width:       Available width in pixels.

    :arg height:      Available height in pixels.

    :arg canvasAxes:  A sequence of ``(xax, yax)`` indices, one for each
                      bitmap in ``canvasBmps``.

    :arg layout:      Either ``'horizontal'``, ``'vertical'``, or ``'grid'``,
                      describing the canvas layout.

    :returns:         A list of ``(width, height)`` tuples, one for each
                      canvas, each specifying the canvas width and height in
                      pixels.
    """

    bounds   = displayCtx.bounds
    axisLens = [bounds.xlen, bounds.ylen, bounds.zlen]

    # Grid layout only makes sense if we're
    # displaying all three canvases
    if layout == 'grid' and len(canvasAxes) <= 2:
        raise ValueError('Grid layout only supports 3 canvases')

    # Distribute the height across canvas heights
    return fsllayout.calcSizes(layout,
                               canvasAxes,
                               axisLens,
                               width,
                               height)


def autocrop(data, bgColour, border=0):
    """Crops the given bitmap image on all sides where the ``bgColour`` is
    the only colour present.

    If the image is completely empty. it is not cropped.

    :arg data:     ``numpy`` array of shape ``(w, h, 4)`` containing the image.
    :arg bgColour: Sequence of length 4 containing the background colour to
                   crop.
    :arg border:   Number of pixels to leave around each side.
    """

    w, h = data.shape[:2]

    low, hiw = 0, w
    loh, hih = 0, h

    while np.all(data[low,     :] == bgColour): low += 1
    while np.all(data[hiw - 1, :] == bgColour): hiw -= 1
    while np.all(data[:, loh]     == bgColour): loh += 1
    while np.all(data[:, hih - 1] == bgColour): hih -= 1

    if low < hiw and loh < hih:
        data = data[low:hiw, loh:hih, :]

        if border > 0:
            w, h, c = data.shape
            new = np.zeros((w + 2 * border, h + 2 * border, c),
                           dtype=data.dtype)
            new[:, :] = bgColour
            new[border:-border, border:-border, :] = data
            data = new

    return data


class MockSliceCanvas(object):
    """Used in place of a :class:`.SliceCanvas`. The :mod:`.parseargs` module
    needs access to ``SliceCanvas`` instances to apply some command line
    options. However, ``render`` calls :func:`.parseargs.applySceneArgs`
    before any ``SliceCanvas`` instances have been created.

    Instances of this class are just used to capture those options, so they
    can later be applied to the real ``SliceCanvas`` instances.

    The following arguments may be applied to this class.
      - ``--xcentre``
      - ``--ycentre``
      - ``--zcentre``
    """
    def __init__(self):
        self.centre = None
    def centreDisplayAt(self, x, y):
        self.centre = x, y


class MockCanvasPanel(object):
    """Used in place of a :class:`.CanvasPanel`. This is used as a container
    for :class:`MockSliceCanvas` instances.
    """
    def __init__(self, ncanvases):
        self.canvases = [MockSliceCanvas() for i in range(ncanvases)]
    def getGLCanvases(self):
        return self.canvases


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# timeseriestoolbar.py - The TimeSeriesToolBar class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`TimeSeriesToolBar`, a toolbar for use with
a :class:`.TimeSeriesPanel`.
"""


import fsleyes_props    as props

import fsleyes.icons    as icons
import fsleyes.strings  as strings
import fsleyes.tooltips as tooltips
import fsleyes.actions  as actions

from . import plottoolbar


class TimeSeriesToolBar(plottoolbar.PlotToolBar):
    """The ``TimeSeriesToolBar`` is a toolbar for use with a
    :class:`.TimeSeriesPanel`. It extends :class:`.PlotToolBar`,
    and adds a few controls specific to the :class:`.TimeSeriesPanel`.
    """

    def __init__(self, parent, overlayList, displayCtx, frame, tsPanel):
        """Create a ``TimeSeriesToolBar``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg frame:       The :class:`.FSLeyesFrame` instance.
        :arg tsPanel:     The :class:`.TimeSeriesPanel` instance.
        """

        plottoolbar.PlotToolBar.__init__(
            self, parent, overlayList, displayCtx, frame, tsPanel)

        togControl = actions.ToggleActionButton(
            'toggleTimeSeriesControl',
            actionKwargs={'floatPane' : True},
            icon=[icons.findImageFile('spannerHighlight24'),
                  icons.findImageFile('spanner24')],
            tooltip=tooltips.actions[tsPanel, 'toggleTimeSeriesControl'])

        togList = actions.ToggleActionButton(
            'togglePlotList',
            actionKwargs={'floatPane' : True},
            icon=[icons.findImageFile('listHighlight24'),
                  icons.findImageFile('list24')],
            tooltip=tooltips.actions[tsPanel, 'togglePlotList'])

        mode = props.Widget('plotMode',
                            labels=strings.choices[     tsPanel, 'plotMode'],
                            tooltip=tooltips.properties[tsPanel, 'plotMode'])

        togControl = props.buildGUI(self, tsPanel, togControl)
        togList    = props.buildGUI(self, tsPanel, togList)
        mode       = props.buildGUI(self, tsPanel, mode)

        lblMode = self.MakeLabelledTool(
            mode, strings.properties[tsPanel, 'plotMode'])

        self.InsertTools([togControl, togList], 0)
        self.AddTool(lblMode)

        nav = [togControl, togList] + self.getCommonNavOrder() + [mode]
        self.setNavOrder(nav)


    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``TimeSeriesToolBar`` is only intended to be added to
        :class:`.TimeSeriesPanel` views.
        """
        from fsleyes.views.timeseriespanel import TimeSeriesPanel
        return [TimeSeriesPanel]



#!/usr/bin/env python
#
# cropimagepanel.py - The CropImagePanel class
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`CropImagePanel` class.

The ``CropImagePanel`` is a a FSLeyes control which is used in conjunction
with the :class:`.OrthoCropProfile`, allowing the user to crop an image.
This module also provides the standalone :func:`loadCropParameters` function,
for loading cropping parameters from a file.
"""


import              os
import os.path   as op
import itertools as it
import              wx
import numpy     as np

import fsl.utils.idle                       as idle
import fsl.data.image                       as fslimage

import fsleyes_props                        as props
import fsleyes_widgets.rangeslider          as rslider
import fsleyes_widgets.utils.status         as status

import fsleyes.controls.controlpanel        as ctrlpanel
import fsleyes.displaycontext               as displaycontext
import fsleyes.strings                      as strings
import fsleyes.actions.copyoverlay          as copyoverlay
import fsleyes.controls.displayspacewarning as dswarning


class CropImagePanel(ctrlpanel.ControlPanel):
    """The ``CropImagePanel`` class is a FSLeyes control for use in an
    :class:`.OrthoPanel`, with the associated :class:`.CropImageProfile`.  It
    contains controls allowing the user to define a cropping box for the
    currently selected overlay (if it is an :class:`.Image`), and "Crop",
    "Load", "Save", and "Cancel" buttons.
    """


    def __init__(self, parent, overlayList, displayCtx, frame, ortho):
        """Create a ``CropImagePanel``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg frame:       The :class:`.FSLeyesFrame` instance.
        :arg ortho:       The :class:`.OrthoPanel` instance.
        """
        ctrlpanel.ControlPanel.__init__(
            self, parent, overlayList, displayCtx, frame)

        profile = ortho.getCurrentProfile()

        self.__ortho   = ortho
        self.__profile = profile
        self.__overlay = None

        self.__cropBoxWidget = props.makeWidget(
            self,
            profile,
            'cropBox',
            showLimits=False,
            labels=['xmin', 'xmax', 'ymin', 'ymax', 'zmin', 'zmax'])

        self.__volumeWidget = rslider.RangeSliderSpinPanel(
            self,
            minValue=0,
            maxValue=1,
            minDistance=1,
            lowLabel='tmin',
            highLabel='tmax',
            style=rslider.RSSP_INTEGER)

        self.__dsWarning = dswarning.DisplaySpaceWarning(
            self,
            overlayList,
            displayCtx,
            frame,
            strings.messages[self, 'dsWarning'],
            'not like overlay',
            'overlay')

        self.__cropLabel       = wx.StaticText(self)
        self.__sizeLabel       = wx.StaticText(self)
        self.__cropButton      = wx.Button(    self, id=wx.ID_OK)
        self.__robustFovButton = wx.Button(    self)
        self.__loadButton      = wx.Button(    self)
        self.__saveButton      = wx.Button(    self)
        self.__cancelButton    = wx.Button(    self, id=wx.ID_CANCEL)

        self.__cropButton     .SetLabel(strings.labels[self, 'crop'])
        self.__robustFovButton.SetLabel(strings.labels[self, 'robustFov'])
        self.__loadButton     .SetLabel(strings.labels[self, 'load'])
        self.__saveButton     .SetLabel(strings.labels[self, 'save'])
        self.__cancelButton   .SetLabel(strings.labels[self, 'cancel'])

        self.__sizer    = wx.BoxSizer(wx.VERTICAL)
        self.__btnSizer = wx.BoxSizer(wx.HORIZONTAL)

        self.__sizer.Add((1, 10))
        self.__sizer.Add(self.__cropLabel,     flag=wx.CENTRE)
        self.__sizer.Add((1, 10))
        self.__sizer.Add(self.__dsWarning,     flag=wx.CENTRE)
        self.__sizer.Add((1, 10), proportion=1)
        self.__sizer.Add(self.__cropBoxWidget, flag=wx.EXPAND)
        self.__sizer.Add(self.__volumeWidget,  flag=wx.EXPAND)
        self.__sizer.Add((1, 10))
        self.__sizer.Add(self.__sizeLabel,     flag=wx.CENTRE, proportion=1)
        self.__sizer.Add((1, 10))
        self.__sizer.Add(self.__btnSizer,      flag=wx.CENTRE)
        self.__sizer.Add((1, 10))

        self.__btnSizer.Add((10, 1),                flag=wx.EXPAND,
                            proportion=1)
        self.__btnSizer.Add(self.__cropButton,      flag=wx.EXPAND)
        self.__btnSizer.Add((10, 1),                flag=wx.EXPAND)
        self.__btnSizer.Add(self.__robustFovButton, flag=wx.EXPAND)
        self.__btnSizer.Add((10, 1),                flag=wx.EXPAND)
        self.__btnSizer.Add(self.__loadButton,      flag=wx.EXPAND)
        self.__btnSizer.Add((10, 1),                flag=wx.EXPAND)
        self.__btnSizer.Add(self.__saveButton,      flag=wx.EXPAND)
        self.__btnSizer.Add((10, 1),                flag=wx.EXPAND)
        self.__btnSizer.Add(self.__cancelButton,    flag=wx.EXPAND)
        self.__btnSizer.Add((10, 1),                flag=wx.EXPAND,
                            proportion=1)

        self.SetSizer(self.__sizer)
        self.SetMinSize(self.__sizer.GetMinSize())
        self.__cropButton.SetDefault()

        self.__cropButton  .Bind(wx.EVT_BUTTON,          self.__onCrop)
        self.__loadButton  .Bind(wx.EVT_BUTTON,          self.__onLoad)
        self.__saveButton  .Bind(wx.EVT_BUTTON,          self.__onSave)
        self.__cancelButton.Bind(wx.EVT_BUTTON,          self.__onCancel)
        self.__volumeWidget.Bind(rslider.EVT_RANGE,      self.__onVolume)
        self.__volumeWidget.Bind(rslider.EVT_LOW_RANGE,  self.__onVolume)
        self.__volumeWidget.Bind(rslider.EVT_HIGH_RANGE, self.__onVolume)

        profile.robustfov.bindToWidget(self,
                                       wx.EVT_BUTTON,
                                       self.__robustFovButton)

        displayCtx .addListener('selectedOverlay',
                                self.name,
                                self.__selectedOverlayChanged)
        overlayList.addListener('overlays',
                                self.name,
                                self.__selectedOverlayChanged)
        profile    .addListener('cropBox',
                                self.name,
                                self.__cropBoxChanged)

        self.__selectedOverlayChanged()
        self.__cropBoxChanged()


    def destroy(self):
        """Must be called when this ``CropImagePanel`` is no longer needed.
        Removes property listeners and clears references.
        """

        profile     = self.__profile
        displayCtx  = self.displayCtx
        overlayList = self.overlayList
        dsWarning   = self.__dsWarning

        profile    .removeListener('cropBox',         self.name)
        displayCtx .removeListener('selectedOverlay', self.name)
        overlayList.removeListener('overlays',        self.name)

        self.__ortho     = None
        self.__profile   = None
        self.__dsWarning = None

        dsWarning.destroy()
        ctrlpanel.ControlPanel.destroy(self)


    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``CropImagePanel`` is only intended to be added to
        :class:`.OrthoPanel` views.
        """
        from fsleyes.views.orthopanel import OrthoPanel
        return [OrthoPanel]


    def __registerOverlay(self, overlay):
        """Called by :meth:`__selectedOverlayChanged`. Registers the
        given overlay.
        """

        self.__overlay = overlay

        display = self.displayCtx.getDisplay(overlay)
        is4D    = overlay.ndim >= 4

        if is4D:
            self.__volumeWidget.SetLimits(0, overlay.shape[3])
            self.__volumeWidget.SetRange( 0, overlay.shape[3])

        self.__volumeWidget.Enable(is4D)

        display.addListener('name', self.name, self.__overlayNameChanged)
        self.__overlayNameChanged()


    def __deregisterOverlay(self):
        """Called by :meth:`__selectedOverlayChanged`. Deregisters the
        current overlay.
        """

        if self.__overlay is None:
            return

        try:
            display = self.displayCtx.getDisplay(self.__overlay)
            display.removeListener('name', self.name)

        except displaycontext.InvalidOverlayError:
            pass

        self.__cropLabel.SetLabel(strings.labels[self, 'image.noImage'])

        self.__overlay = None


    def __overlayNameChanged(self, *a):
        """Called when the :attr:`.Display.name` of the currently selected
        overlay changes. Updates the name label.
        """

        display = self.displayCtx.getDisplay(self.__overlay)
        label   = strings.labels[self, 'image']
        label   = label.format(display.name)
        self.__cropLabel.SetLabel(label)


    def __selectedOverlayChanged(self, *a):
        """Called when the :attr:`.DisplayContext.selectedOverlay` changes.
        Updates labels appropriately.
        """

        displayCtx = self.displayCtx
        overlay    = displayCtx.getSelectedOverlay()

        if overlay is self.__overlay:
            return

        self.__deregisterOverlay()

        if not isinstance(overlay, fslimage.Image):
            self.Disable()
        else:
            self.Enable()
            self.__registerOverlay(overlay)


    def __updateSizeLabel(self):
        """Called by the crop region and volume widget event handlers. Updates
        a label which displays the current crop region size.
        """

        overlay = self.__overlay
        profile = self.__profile
        xlen    = profile.cropBox.xlen
        ylen    = profile.cropBox.ylen
        zlen    = profile.cropBox.zlen
        tlo     = self.__volumeWidget.GetLow()
        thi     = self.__volumeWidget.GetHigh()
        tlen    = thi - tlo

        if overlay.ndim >= 4:
            label = strings.labels[self, 'cropSize4d']
            label = label.format(xlen, ylen, zlen, tlen)
        else:
            label = strings.labels[self, 'cropSize3d']
            label = label.format(xlen, ylen, zlen)

        self.__sizeLabel.SetLabel(label)


    def __cropBoxChanged(self, *a):
        """Called when the :attr:`.OrthoCropProfile.cropBox` changes.
        Updates labels appropriately.
        """
        self.__updateSizeLabel()


    def __onVolume(self, ev):
        """Called when the user changes the volume limit, for 4D images.
        Updates the label which displays the crop region size.
        """
        self.__updateSizeLabel()


    def __onLoad(self, ev):
        """Called when the Save button is pushed. Prompts the user to select
        a file to load crop parameters from.
        """

        overlay  = self.__overlay
        cropBox  = self.__profile.cropBox
        fileName = '{}_crop.txt'.format(overlay.name)

        if overlay.dataSource is not None:
            dirName = op.dirname(overlay.dataSource)
        else:
            dirName = os.getcwd()

        if not op.exists(op.join(dirName, fileName)):
            fileName = ''

        dlg = wx.FileDialog(
            self,
            defaultDir=dirName,
            defaultFile=fileName,
            message=strings.messages[self, 'saveCrop'],
            style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST)

        if dlg.ShowModal() != wx.ID_OK:
            return

        filePath = dlg.GetPath()
        errTitle = strings.titles[  self, 'loadError']
        errMsg   = strings.messages[self, 'loadError']

        with status.reportIfError(errTitle, errMsg, raiseError=False):

            params     = loadCropParameters(filePath, overlay)
            cropBox[:] = params[:6]

            if overlay.ndim >= 4:
                tlo, thi = params[6:]
                self.__volumeWidget.SetLow(tlo)
                self.__volumeWidget.SetHigh(thi)


    def __onSave(self, ev):
        """Called when the Save button is pushed. Saves the current crop
        parameters to a text file.
        """

        overlay  = self.__overlay
        cropBox  = self.__profile.cropBox
        fileName = '{}_crop.txt'.format(overlay.name)

        if overlay.dataSource is not None:
            dirName = op.dirname(overlay.dataSource)
        else:
            dirName = os.getcwd()

        dlg = wx.FileDialog(
            self,
            defaultDir=dirName,
            defaultFile=fileName,
            message=strings.messages[self, 'saveCrop'],
            style=wx.FD_SAVE)

        if dlg.ShowModal() != wx.ID_OK:
            return

        filePath = dlg.GetPath()

        # The crop parameters are saved
        # in a fslroi-compatible manner.
        params = [cropBox.xlo,
                  cropBox.xhi - cropBox.xlo,
                  cropBox.ylo,
                  cropBox.yhi - cropBox.ylo,
                  cropBox.zlo,
                  cropBox.zhi - cropBox.zlo]

        if overlay.ndim >= 4:
            tlo     = self.__volumeWidget.GetLow()
            thi     = self.__volumeWidget.GetHigh()
            params.extend((tlo, thi - tlo))

        errTitle = strings.titles[  self, 'saveError']
        errMsg   = strings.messages[self, 'saveError']

        with status.reportIfError(errTitle, errMsg, raiseError=False):
            np.savetxt(filePath, [params], fmt='%i')


    def __onCancel(self, ev=None):
        """Called when the Cancel button is pushed. Calls
        :meth:`.OrthoPanel.toggleCropMode` - this will result in
        this ``CropImagePanel`` being destroyed.

        This method is also called programmatically from the :meth:`__onCrop`
        method after the image is cropped.
        """

        # Do asynchronously, because we don't want
        # this CropImagePanel being destroyed from
        # its own event handler.
        idle.idle(self.__ortho.toggleCropMode)


    def __onCrop(self, ev):
        """Crops the selected image. This is done via a call to
        :func:`.copyoverlay.copyImage`. Also calls :meth:`__onCancel`,
        to finish cropping.
        """

        overlayList = self.overlayList
        displayCtx  = self.displayCtx
        overlay     = displayCtx.getSelectedOverlay()
        display     = displayCtx.getDisplay(overlay)
        name        = '{}_roi'.format(display.name)
        cropBox     = self.__profile.cropBox
        roi         = [cropBox.x,
                       cropBox.y,
                       cropBox.z]

        if overlay.ndim >= 4:
            roi.append(self.__volumeWidget.GetRange())

        copyoverlay.copyImage(
            overlayList,
            displayCtx,
            overlay,
            createMask=False,
            copy4D=True,
            copyDisplay=True,
            name=name,
            roi=roi)

        self.__onCancel()


def loadCropParameters(filename, overlay):
    """Load in crop values from a text file assumed to contain ``fslroi``-
    compatible parameters. Any parameters which may be passed to ``fslroi``
    are accepted::

        fslroi in out tmin tlen
        fslroi in out xmin xlen ymin ylen zmin zlen
        fslroi in out xmin xlen ymin ylen zmin zlen tmin tlen

    Any of the ``len`` parameters may be equal to -1, in which case it is
    interpreted as continuing from the low index

    :arg filename: File to load crop parameters from.
    :arg overlay:  An :class:`.Image` which is the cropping target.
    :returns:      A sequence of ``lo, hi`` crop parameters.
    """

    is4D   = overlay.ndim >= 4
    shape  = overlay.shape[:4]
    params = list(np.loadtxt(filename).flatten())

    if len(params) not in (2, 6, 8):
        raise ValueError('File contains the wrong number of crop parameters')
    if len(params) in (2, 8) and not is4D:
        raise ValueError('File contains the wrong number of crop parameters')

    if len(params) == 2:
        params = [0, -1, 0, -1, 0, -1] + params

    if is4D and len(params) == 6:
        params = params + [0, -1]

    los = []
    his = []

    for dim in range(len(shape)):

        dlo  = params[dim * 2]
        dlen = params[dim * 2 + 1]

        if dlen == -1:
            dlen = shape[dim] - dlo

        dhi = dlo + dlen

        los.append(dlo)
        his.append(dhi)

    for lo, hi, lim in zip(los, his, shape):
        if lo < 0 or hi > lim:
            raise ValueError('Crop parameters are out of bounds for image '
                             'shape ({} < 0 or {} > {}'.format(lo, hi, lim))

    return list(it.chain(*zip(los, his)))



#!/usr/bin/env python
#
# orthopanel.py - The OrthoPanel class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`OrthoPanel` class, which displays a 2D
view of 3D overlays.

A couple of other classes are provided for convenience:

.. autosummary::
   :nosignatures:

   OrthoFrame
   OrthoDialog
"""


import logging

import wx

import fsl.data.image                          as fslimage
import fsleyes_widgets.utils.layout            as fsllayout

import fsleyes.strings                         as strings
import fsleyes.gl                              as fslgl
import fsleyes.actions                         as actions
import fsleyes.editor                          as editor
import fsleyes.gl.ortholabels                  as ortholabels
import fsleyes.gl.wxglslicecanvas              as slicecanvas
import fsleyes.controls.cropimagepanel         as cropimagepanel
import fsleyes.controls.edittransformpanel     as edittransformpanel
import fsleyes.controls.orthotoolbar           as orthotoolbar
import fsleyes.controls.orthoedittoolbar       as orthoedittoolbar
import fsleyes.controls.orthoeditactiontoolbar as orthoeditactiontoolbar
import fsleyes.controls.orthoeditsettingspanel as orthoeditsettingspanel
import fsleyes.displaycontext.orthoopts        as orthoopts
from . import                                     canvaspanel


log = logging.getLogger(__name__)


class OrthoPanel(canvaspanel.CanvasPanel):
    """The ``OrthoPanel`` class is a *FSLeyes view* which displays a 2D view
    of 3D overlays.  The ``OrthoPanel`` is the primary point of user
    interaction in *FSLeyes*.


    **Overview**


    An ``OrthoPanel`` contains three :class:`.SliceCanvas` panels, each of
    which provide a 2D view of the overlays in the :class:`.OverlayList` along
    one axis. These ``SliceCanvas`` instances can be accessed through the
    :meth:`getXCanvas`, :meth:`getYCanvas`, :meth:`getZCanvas`, and
    :meth:`getGLCanvases` methods.


    An ``OrthoPanel`` looks something like this:


    .. image:: images/orthopanel.png
       :scale: 50%
       :align: center


    **Anatomical labels**


    The ``OrthoPanel`` creates an :class:`.OrthoLabels` instance, which
    manages the display of anatomical orientation labels on each of the
    three :class:`.SliceCanvas` instances.


    **Display**


    The display of an ``OrthoPanel`` can be configured through all of the
    settings provided by the :class:`.OrthoOpts` class. The ``OrthoOpts``
    instance for a given ``OrthoPanel`` can be accessed via the
    :meth:`.CanvasPanel.sceneOpts` method.


    **Interaction**


    The following interaction profiles are defined for use with the
    ``OrthoPanel`` (see the :class:`.ViewPanel` for an overview of
    *profiles*):

    ======== =========================================================
    ``view`` Viewing/navigation, using the :class:`.OrthoViewProfile`.

    ``edit`` Simple editing of :class:`.Image` overlays, using the
             :class:`.OrthoEditProfile` (see also the
             :mod:`~fsleyes.editor` package).

    ``crop`` Allows the user to crop an ``Image`` overlay.
    ======== =========================================================


    **Actions and control panels**


    The ``OrthoPanel`` adds a few extra actions to those provided by the
    :class:`.CanvasPanel` class:

    .. autosummary::
       :nosignatures:

       toggleEditMode
       toggleCropMode
       toggleEditTransformPanel
       toggleEditPanel
       toggleOrthoToolBar
       resetDisplay
       centreCursor
       centreCursorWorld
       toggleCursor
       toggleLabels
       toggleXCanvas
       toggleYCanvas
       toggleZCanvas
    """


    def __init__(self, parent, overlayList, displayCtx, frame):
        """Create an ``OrthoPanel``.

        :arg parent:      The :mod:`wx` parent.
        :arg overlayList: An :class:`.OverlayList` instance.
        :arg displayCtx:  A :class:`.DisplayContext` instance.
        :arg displayCtx:  A :class:`.FSLeyesFrame` instance.
        """

        sceneOpts = orthoopts.OrthoOpts(self)

        canvaspanel.CanvasPanel.__init__(self,
                                         parent,
                                         overlayList,
                                         displayCtx,
                                         frame,
                                         sceneOpts)

        name         = self.name
        contentPanel = self.contentPanel

        # The canvases themselves - each one displays
        # a slice along each of the three world axes.
        # They are laid out in __refreshLayout
        self.__canvasSizer = None
        self.__xcanvas = slicecanvas.WXGLSliceCanvas(contentPanel,
                                                     overlayList,
                                                     displayCtx,
                                                     zax=0)
        self.__ycanvas = slicecanvas.WXGLSliceCanvas(contentPanel,
                                                     overlayList,
                                                     displayCtx,
                                                     zax=1)
        self.__zcanvas = slicecanvas.WXGLSliceCanvas(contentPanel,
                                                     overlayList,
                                                     displayCtx,
                                                     zax=2)

        self.__labelMgr = ortholabels.OrthoLabels(
            overlayList,
            displayCtx,
            sceneOpts,
            self.__xcanvas,
            self.__ycanvas,
            self.__zcanvas)

        # If an edit menu is added when in
        # 'edit' profile (see __profileChanged),
        # its name is stored here.
        self.__editMenuTitle = None

        xopts = self.__xcanvas.opts
        yopts = self.__ycanvas.opts
        zopts = self.__zcanvas.opts

        xopts.bindProps('pos', displayCtx, 'location')
        yopts.bindProps('pos', displayCtx, 'location')
        zopts.bindProps('pos', displayCtx, 'location')

        xopts.bindProps('showCursor',   sceneOpts)
        yopts.bindProps('showCursor',   sceneOpts)
        zopts.bindProps('showCursor',   sceneOpts)

        xopts.bindProps('cursorGap',    sceneOpts)
        yopts.bindProps('cursorGap',    sceneOpts)
        zopts.bindProps('cursorGap',    sceneOpts)

        xopts.bindProps('bgColour',     sceneOpts)
        yopts.bindProps('bgColour',     sceneOpts)
        zopts.bindProps('bgColour',     sceneOpts)

        xopts.bindProps('cursorColour', sceneOpts)
        yopts.bindProps('cursorColour', sceneOpts)
        zopts.bindProps('cursorColour', sceneOpts)

        xopts.bindProps('zoom',         sceneOpts, 'xzoom')
        yopts.bindProps('zoom',         sceneOpts, 'yzoom')
        zopts.bindProps('zoom',         sceneOpts, 'zzoom')

        xopts.bindProps('renderMode',   sceneOpts)
        yopts.bindProps('renderMode',   sceneOpts)
        zopts.bindProps('renderMode',   sceneOpts)

        xopts.bindProps('highDpi',      sceneOpts)
        yopts.bindProps('highDpi',      sceneOpts)
        zopts.bindProps('highDpi',      sceneOpts)

        # Callbacks for ortho panel layout options
        sceneOpts.addListener('layout', name, self.__refreshLayout)

        self.toggleCursor .bindProps('toggled', sceneOpts, 'showCursor')
        self.toggleLabels .bindProps('toggled', sceneOpts, 'showLabels')
        self.toggleXCanvas.bindProps('toggled', sceneOpts, 'showXCanvas')
        self.toggleYCanvas.bindProps('toggled', sceneOpts, 'showYCanvas')
        self.toggleZCanvas.bindProps('toggled', sceneOpts, 'showZCanvas')

        # Callbacks for overlay list/selected overlay changes
        overlayList.addListener('overlays',
                                name,
                                self.__overlayListChanged)
        displayCtx .addListener('bounds',
                                name,
                                self.__refreshLayout)
        displayCtx .addListener('displaySpace',
                                name,
                                self.__radioOrientationChanged)
        displayCtx .addListener('radioOrientation',
                                name,
                                self.__radioOrientationChanged)
        displayCtx .addListener('selectedOverlay',
                                name,
                                self.__overlayListChanged)

        # Callbacks for toggling x/y/z canvas display
        sceneOpts.addListener('showXCanvas', name, self.__toggleCanvas)
        sceneOpts.addListener('showYCanvas', name, self.__toggleCanvas)
        sceneOpts.addListener('showZCanvas', name, self.__toggleCanvas)

        # Callbacks which just need to refresh
        def refresh(*a):
            self.Refresh()

        sceneOpts.addListener('labelSize',   name, refresh, weak=False)
        sceneOpts.addListener('fgColour',    name, refresh, weak=False)
        sceneOpts.addListener('showLabels',  name, refresh, weak=False)

        self.addListener('profile', name, self.__profileChanged)

        from fsleyes.actions.correlate import PearsonCorrelateAction

        self.__pCorrAction = PearsonCorrelateAction(
            self.overlayList,
            self.displayCtx,
            self)

        self.pearsonCorrelation.bindProps('enabled', self.__pCorrAction)

        # Call the __onResize method to refresh
        # the slice canvases when the canvas
        # panel is resized, so aspect ratio
        # is maintained
        contentPanel.Bind(wx.EVT_SIZE, self.__onResize)

        # Initialise the panel
        self.__radioOrientationChanged()
        self.__refreshLayout(refresh=False)
        self.__overlayListChanged()
        self.centrePanelLayout()
        self.initProfile()


    def destroy(self):
        """Must be called when this ``OrthoPanel`` is closed.

        Removes listeners from the :class:`.DisplayContext` and
        :class:`.OverlayList` instances, destroys each of the three
        :class:`.SliceCanvas` panels, and calls :meth:`.CanvasPanel.destroy`.
        """

        sceneOpts    = self.sceneOpts
        contentPanel = self.contentPanel

        sceneOpts       .removeListener('showXCanvas',      self.name)
        sceneOpts       .removeListener('showYCanvas',      self.name)
        sceneOpts       .removeListener('showZCanvas',      self.name)
        sceneOpts       .removeListener('labelSize',        self.name)
        sceneOpts       .removeListener('fgColour',         self.name)
        sceneOpts       .removeListener('showLabels',       self.name)
        self.displayCtx .removeListener('location',         self.name)
        self.displayCtx .removeListener('bounds',           self.name)
        self.displayCtx .removeListener('selectedOverlay',  self.name)
        self.displayCtx .removeListener('displaySpace',     self.name)
        self.displayCtx .removeListener('radioOrientation', self.name)
        self.overlayList.removeListener('overlays',         self.name)

        self.__labelMgr.destroy()
        self.__xcanvas.destroy()
        self.__ycanvas.destroy()
        self.__zcanvas.destroy()
        self.__removeEditMenu()
        self.__pCorrAction.destroy()

        contentPanel.Unbind(wx.EVT_SIZE)

        self.__xcanvas     = None
        self.__ycanvas     = None
        self.__zcanvas     = None
        self.__labelMgr    = None
        self.__pCorrAction = None

        canvaspanel.CanvasPanel.destroy(self)


    @actions.toggleControlAction(orthotoolbar.OrthoToolBar)
    def toggleOrthoToolBar(self):
        """Shows/hides an :class:`.OrthoToolBar`. See
        :meth:`.ViewPanel.togglePanel`.
        """
        self.togglePanel(orthotoolbar.OrthoToolBar, ortho=self)


    @actions.toggleControlAction(orthoedittoolbar.OrthoEditToolBar)
    def toggleEditMode(self):
        """Toggles the :attr:`.ViewPanel.profile` between ``'view'`` and
        ``'edit'``. See :meth:`__profileChanged`.
        """

        if self.profile == 'view': self.profile = 'edit'
        else:                      self.profile = 'view'


    @actions.toggleControlAction(cropimagepanel.CropImagePanel)
    def toggleCropMode(self):
        """Toggles the :attr:`.ViewPanel.profile` between ``'view'`` and
        ``'crop'``. See :meth:`__profileChanged`.
        """

        if self.profile == 'view': self.profile = 'crop'
        else:                      self.profile = 'view'


    @actions.toggleControlAction(edittransformpanel.EditTransformPanel)
    def toggleEditTransformPanel(self):
        """Shows/hides an :class:`.EditTransformPanel`. See
        :meth:`.ViewPanel.togglePanel`.
        """

        self.togglePanel(edittransformpanel.EditTransformPanel,
                         floatPane=True,
                         floatOnly=True,
                         closeable=False,
                         ortho=self)


    @actions.toggleControlAction(orthoeditsettingspanel.OrthoEditSettingsPanel)
    def toggleEditPanel(self, floatPane=False):
        """Shows/hides an :class:`.OrthoEditSettingsPanel`. See
        :meth:`.ViewPanel.togglePanel`.
        """
        self.togglePanel(orthoeditsettingspanel.OrthoEditSettingsPanel,
                         ortho=self,
                         floatPane=floatPane)


    @actions.action
    def pearsonCorrelation(self):
        """Executes a :class:`.PearsonCorrelateAction`. """
        self.__pCorrAction()


    @actions.action
    def resetDisplay(self):
        """Calls :meth:`.OrthoViewProfile.resetDisplay`. """
        self.getCurrentProfile().resetDisplay()


    @actions.action
    def centreCursor(self):
        """Calls :meth:`.OrthoViewProfile.centreCursor`. """
        self.getCurrentProfile().centreCursor()


    @actions.action
    def centreCursorWorld(self):
        """Calls :meth:`.OrthoViewProfile.centreCursorWorld`. """
        self.getCurrentProfile().centreCursorWorld()


    @actions.toggleAction
    def toggleCursor(self):
        """Toggles the value of :attr:`.SceneOpts.showCursor`. """
        # The state of this action gets bound to
        # the showCursor attribute in __init__
        pass


    @actions.toggleAction
    def toggleLabels(self):
        """Toggles the value of :attr:`.OrthoOpts.showLabels`. """
        # The state of this action gets bound to
        # the showLabels attribute in __init__
        pass


    @actions.toggleAction
    def toggleXCanvas(self):
        """Toggles the value of :attr:`.OrthoOpts.showXCanvas`. """
        # The state of this action gets bound to
        # the showXCanvas attribute in __init__
        pass


    @actions.toggleAction
    def toggleYCanvas(self):
        """Toggles the value of :attr:`.OrthoOpts.showYCanvas`. """
        # The state of this action gets bound to
        # the showYCanvas attribute in __init__
        pass


    @actions.toggleAction
    def toggleZCanvas(self):
        """Toggles the value of :attr:`.OrthoOpts.showZCanvas`. """
        # The state of this action gets bound to
        # the showZCanvas attribute in __init__
        pass


    def getActions(self):
        """Overrides :meth:`.ActionProvider.getActions`. Returns all of the
        :mod:`.actions` that are defined on this ``OrthoPanel``.
        """
        actionz = [self.screenshot,
                   self.movieGif,
                   self.showCommandLineArgs,
                   self.applyCommandLineArgs,
                   None,
                   self.toggleMovieMode,
                   self.toggleDisplaySync,
                   None,
                   self.resetDisplay,
                   self.centreCursor,
                   self.centreCursorWorld,
                   None,
                   self.toggleLabels,
                   self.toggleCursor,
                   self.toggleXCanvas,
                   self.toggleYCanvas,
                   self.toggleZCanvas,
                   None,
                   self.toggleOverlayList,
                   self.toggleLocationPanel,
                   self.toggleOverlayInfo,
                   self.toggleDisplayPanel,
                   self.toggleCanvasSettingsPanel,
                   self.toggleAtlasPanel,
                   self.toggleDisplayToolBar,
                   self.toggleOrthoToolBar,
                   self.toggleFileTreePanel,
                   self.toggleLookupTablePanel,
                   self.toggleClusterPanel,
                   self.toggleClassificationPanel,
                   self.removeAllPanels]

        def makeTuples(actionz):

            tuples = []

            for a in actionz:
                if isinstance(a, actions.Action):
                    tuples.append((a.__name__, a))

                elif isinstance(a, tuple):
                    tuples.append((a[0], makeTuples(a[1])))

                elif a is None:
                    tuples.append((None, None))

            return tuples

        return makeTuples(actionz)


    def getTools(self):
        """Returns a list of methods to be added to the ``FSLeyesFrame`` Tools menu
        for ``OrthoPanel`` views.
        """
        return [self.toggleEditMode,
                self.toggleCropMode,
                self.toggleEditTransformPanel,
                self.pearsonCorrelation]


    def getGLCanvases(self):
        """Returns all of the :class:`.SliceCanvas` instances contained
        within this ``OrthoPanel``.
        """
        return [self.__xcanvas, self.__ycanvas, self.__zcanvas]


    def getXCanvas(self):
        """Returns the :class:`.SliceCanvas` instance displaying the X axis.
        """
        return self.__xcanvas


    def getYCanvas(self):
        """Returns the :class:`.SliceCanvas` instance displaying the Y axis.
        """
        return self.__ycanvas


    def getZCanvas(self):
        """Returns the :class:`.SliceCanvas` instance displaying the Z axis.
        """
        return self.__zcanvas


    def __profileChanged(self, *a):
        """Called when the :attr:`.ViewPanel.profile` changes. If ``'edit'``
        mode has been enabled, :class:`.OrthEditToolBar` and
        :class:`.OrthEditActionToolBar` toolbars are added as control panels,
        and an "edit" menu is added to the :class:`.FSLeyesFrame` (if there
        is one).
        """

        CropImagePanel         = cropimagepanel.CropImagePanel
        OrthoEditToolBar       = orthoedittoolbar.OrthoEditToolBar
        OrthoEditActionToolBar = orthoeditactiontoolbar.OrthoEditActionToolBar
        OrthoEditSettingsPanel = orthoeditsettingspanel.OrthoEditSettingsPanel

        cropPanelOpen          = self.isPanelOpen(CropImagePanel)
        editToolBarOpen        = self.isPanelOpen(OrthoEditToolBar)
        editActionToolBarOpen  = self.isPanelOpen(OrthoEditActionToolBar)
        editPanelOpen          = self.isPanelOpen(OrthoEditSettingsPanel)

        inEdit                 = self.profile == 'edit'
        inCrop                 = self.profile == 'crop'

        # Toggle toolbars if they are open but should
        # be closed, or closed but should be open
        if (not editToolBarOpen) and      inEdit or \
                editToolBarOpen  and (not inEdit):
            self.togglePanel(orthoedittoolbar.OrthoEditToolBar, ortho=self)

        if (not editActionToolBarOpen) and      inEdit or \
                editActionToolBarOpen  and (not inEdit):
            self.togglePanel(orthoeditactiontoolbar.OrthoEditActionToolBar,
                             ortho=self,
                             location=wx.LEFT)

        if (not cropPanelOpen) and      inCrop or \
                cropPanelOpen  and (not inCrop):
            self.togglePanel(cropimagepanel.CropImagePanel,
                             ortho=self,
                             floatPane=True,
                             floatOnly=True,
                             closeable=False,
                             floatPos=(0.85, 0.3))

        # Don't open edit panel by default,
        # but close it when we leave edit mode
        if editPanelOpen and (not inEdit):
            self.togglePanel(orthoeditsettingspanel.OrthoEditSettingsPanel)

        # It's unlikely, but an OrthoPanel might be
        # created without a ref to a FSLeyesFrame.
        if self.frame is not None:
            if inEdit: self.__addEditMenu()
            else:      self.__removeEditMenu()


    def __addEditMenu(self):
        """Called by :meth:`__profleChanged` when the
        :attr:`.ViewPanel.profile` is changed to ``'edit'``. Adds a
        menu to the :class:`.FSLeyesFrame`.
        """

        frame    = self.frame
        menuName = strings.labels[self, 'editMenu']
        menuName = menuName.format(frame.getViewPanelID(self))
        menuBar  = frame.GetMenuBar()
        profile  = self.getCurrentProfile()
        idx      = menuBar.FindMenu(menuName)

        if  idx == wx.NOT_FOUND: editMenu = None
        else:                    editMenu = menuBar.GetMenu(idx)

        if editMenu is not None:
            return

        self.__editMenuTitle = menuName

        editMenu = wx.Menu()

        menuBar.Append(editMenu, menuName)

        actionz = ['undo',
                   'redo',
                   'createMask',
                   'clearSelection',
                   'fillSelection',
                   'eraseSelection',
                   'copySelection',
                   'pasteSelection',
                   'invertSelection']

        frame.populateMenu(editMenu,
                           profile,
                           actionz,
                           ignoreFocus=True)

        # Add a 'close' option too, but run it
        # on the idle loop, as its execution will
        # cause the owning menu to be destroyed.
        frame.populateMenu(editMenu,
                           self,
                           [None, 'toggleEditMode'],
                           ignoreFocus=True,
                           runOnIdle=True)



    def __removeEditMenu(self):
        """Called by :meth:`__profleChanged` when the
        :attr:`.ViewPanel.profile` is changed from ``'edit'``. If an edit
        menut has previously been added to the :class:`.FSLeyesFrame`, it
        is removed.
        """

        if self.__editMenuTitle is None:
            return

        frame        = self.frame
        editMenuName = self.__editMenuTitle
        menuBar      = frame.GetMenuBar()
        idx          = menuBar.FindMenu(editMenuName)

        self.__editMenuTitle = None

        if  idx == wx.NOT_FOUND:
            return

        editMenu = menuBar.GetMenu(idx)

        menuBar.Remove(idx)
        wx.CallAfter(editMenu.Destroy)


    def __toggleCanvas(self, *a):
        """Called when any of the :attr:`.OrthoOpts.showXCanvas`,
        :attr:`.OrthoOpts.showYCanvas`, or :attr:`.OrthoOpts.showZCanvas`
        properties are changed.

        Shows/hides each of the :class:`.SliceCanvas` panels accordingly.
        """

        opts     = self.sceneOpts
        canvases = [self.__xcanvas,   self.__ycanvas,   self.__zcanvas]
        shows    = [opts.showXCanvas, opts.showYCanvas, opts.showZCanvas]

        for canvas, show in zip(canvases, shows):

            # See WXGLSliceCanvas.Show for
            # details of a horrible bug, and
            # equally horrible workaround..
            canvas.Show(show)

        # If layout == grid, then the actual
        # layout may be different depending
        # on how many canvases are displayed
        if opts.layout == 'grid':
            self.__refreshLayout()

        self.PostSizeEvent()


    def __radioOrientationChanged(self, *a):
        """Called when the :attr:`.DisplayContext.radioOrientation` or
        :attr:`.DisplayContext.displaySpace` property changes. Figures out if
        the left-right canvas axes need to be flipped, and does so if
        necessary.
        """

        if len(self.overlayList) == 0:
            return

        inRadio = self.displayCtx.displaySpaceIsRadiological()
        flip    = self.displayCtx.radioOrientation != inRadio

        self.__ycanvas.opts.invertX = flip
        self.__zcanvas.opts.invertX = flip


    def __overlayListChanged(self, *a):
        """Called when the :class:`.OverlayList` or
        :attr:`.DisplayContext.selectedOverlay` is changed. Enables/disables
        various action methods based on the currently selected overlay.
        """

        # Disable actions that need an overlay
        haveOverlays = len(self.overlayList) > 0
        selOverlay   = self.displayCtx.getSelectedOverlay()

        if selOverlay is not None:

            display    = self.displayCtx.getDisplay(selOverlay)
            isImage    = isinstance(selOverlay, fslimage.Image) and \
                             display.overlayType in ('volume',
                                                     'mask',
                                                     'label',
                                                     'mip')
            isEditable = editor.isEditable(selOverlay, self.displayCtx)
        else:
            isImage    = False
            isEditable = False

        self.resetDisplay            .enabled = haveOverlays
        self.centreCursor            .enabled = haveOverlays
        self.centreCursorWorld       .enabled = haveOverlays
        self.toggleEditMode          .enabled = isEditable
        self.toggleEditTransformPanel.enabled = isImage
        self.toggleCropMode          .enabled = isImage

        # Kill edit mode if a non-
        # image has been selected
        if (self.profile == 'edit') and (not isEditable):
            self.profile = 'view'


    def __onResize(self, ev):
        """Called whenever the panel is resized. Makes sure that the
        :class:`.SliceCanvas` panels  and :class:`.Text` annotations
        are drawn correctly.
        """
        ev.Skip()
        self.__calcCanvasSizes()


    def __calcCanvasSizes(self, *a):
        """Sets the size for each displayed :class:`.SliceCanvas`.

        The minimum/maximum size of each canvas is fixed so that they are
        scaled proportionally to each other, thus preserving the aspect ratio.
        The :mod:~fsl.utils.layout` module is used to perform the canvas size
        calculation.
        """

        opts   = self.sceneOpts
        layout = opts.layout

        width, height = self.contentPanel.GetClientSize().Get()

        show     = [opts.showXCanvas,  opts.showYCanvas,  opts.showZCanvas]
        canvases = [self.__xcanvas,    self.__ycanvas,    self.__zcanvas]

        if width == 0 or height == 0:  return
        if len(self.overlayList) == 0: return
        if not any(show):              return

        canvases = [c for (c, s) in zip(canvases, show) if s]

        # Grid layout with 2 or less canvases displayed
        # is identical to horizontal layout
        if layout == 'grid' and len(canvases) <= 2:
            layout = 'horizontal'

        # Grid layout canvas
        # order is YXZ
        if layout == 'grid':
            canvases = [canvases[1], canvases[0], canvases[2]]

        # Distribute the available width/height
        # to each of the displayed canvases -
        # fsleyes_widgets.utils.layout (a.k.a.
        # fsllayout) provides functions to do
        # this for us
        canvasaxes = [(c.opts.xax, c.opts.yax) for c in canvases]
        axisLens   = [self.displayCtx.bounds.xlen,
                      self.displayCtx.bounds.ylen,
                      self.displayCtx.bounds.zlen]

        sizes = fsllayout.calcSizes(layout,
                                    canvasaxes,
                                    axisLens,
                                    width,
                                    height)

        for canvas, size in zip(canvases, sizes):
            canvas.SetMinSize((round(size[0]), round(size[1])))
            canvas.SetMaxSize((round(size[0]), round(size[1])))


    def __refreshLayout(self, *a, **kwa):
        """Called when the :attr:`.OrthoOpts.layout` property changes, or the
        canvas layout needs to be refreshed. Updates the layout accordingly.

        :arg refresh: Must be passed as a keyword argument. If ``True`` (the
                      default), this ``OrthoPanel`` is refreshed.
        """

        refresh = kwa.pop('refresh', True)

        opts   = self.sceneOpts
        layout = opts.layout

        # We lay out all canvases, even
        # the ones that are not shown.
        canvases  = [self.__xcanvas,   self.__ycanvas,   self.__zcanvas]
        shows     = [opts.showXCanvas, opts.showYCanvas, opts.showZCanvas]
        nCanvases = sum(shows)

        # For the grid layout if only one or two
        # canvases are being displayed, the layout
        # is equivalent to a horizontal layout.
        if layout == 'grid' and nCanvases <= 2:
            layout = 'horizontal'

        # For horizontal/vertical layout,
        # the canvas layout is:
        #
        #   | X/sagittal | Y/coronal | Z/axial |
        #
        # But for grid layout, the canvas
        # layout is:
        #
        #   | Y/coronal | X/sagittal |
        #   | Z/axial   |            |
        #
        if layout == 'grid':
            canvases = [self.__ycanvas, self.__xcanvas, self.__zcanvas]

        # Regardless of the layout, we use a
        # FlexGridSizer with varying numbers
        # of rows/columns, depending upon the
        # layout strategy
        if   layout == 'horizontal': nrows, ncols = 1, 3
        elif layout == 'vertical':   nrows, ncols = 3, 1
        elif layout == 'grid':       nrows, ncols = 2, 2

        if self.__canvasSizer is not None:
            self.__canvasSizer.Clear()
            self.__canvasSizer = None

        self.__canvasSizer = wx.FlexGridSizer(nrows, ncols, 0, 0)

        # The rows/columns that contain
        # canvases must also be growable
        for row in range(nrows): self.__canvasSizer.AddGrowableRow(row)
        for col in range(ncols): self.__canvasSizer.AddGrowableCol(col)

        # For grid layout, the last cell is filled with empty space
        if layout == 'grid':
            canvases.append((0, 0))

        # Add all those widgets to the grid sizer
        for c in canvases:
            self.__canvasSizer.Add(c, flag=wx.EXPAND)

        self.contentPanel.SetSizer(self.__canvasSizer)

        # Calculate/ adjust the appropriate sizes
        # for each canvas, such that they are scaled
        # appropriately relative to each other, and
        # the displayed world space aspect ratio is
        # maintained
        self.__calcCanvasSizes()

        # When in grid layout, flip the horizontal axis
        # of the X canvas (assumed to be A/P), to force
        # third angle orthographic projection.
        self.__xcanvas.opts.invertX = layout == 'grid'

        if refresh:
            self.Layout()
            self.contentPanel.Layout()
            self.Refresh()


class OrthoFrame(wx.Frame):
    """Convenience class for displaying an :class:`OrthoPanel` in a
    standalone frame.
    """

    def __init__(self, parent, overlayList, displayCtx, title=None):
        """Create an ``OrthoFrame``.

        :arg parent:      A :mod:`wx` parent object.

        :arg overlayList: An :class:`.OverlayList` instance.

        :arg displayCtx:  A :class:`.DisplayContext` instance.

        :arg title:       Dialog title.
        """

        wx.Frame.__init__(self, parent, title=title)

        fslgl.getGLContext()
        fslgl.bootstrap()

        self.panel = OrthoPanel(self, overlayList, displayCtx, None)
        self.Layout()


class OrthoDialog(wx.Dialog):
    """Convenience class for displaying an :class:`OrthoPanel` in a (possibly
    modal) dialog window.
    """

    def __init__(self,
                 parent,
                 overlayList,
                 displayCtx,
                 title=None,
                 style=None):
        """Create an ``OrthoDialog``.

        :arg parent:      A :mod:`wx` parent object.

        :arg overlayList: An :class:`.OverlayList` instance.

        :arg displayCtx:  A :class:`.DisplayContext` instance.

        :arg title:       Dialog title.

        :arg style:       Dialog style - defaults to
                          ``wx.DEFAULT_DIALOG_STYLE``.
        """

        if style is None:
            style = wx.DEFAULT_DIALOG_STYLE

        wx.Dialog.__init__(self, parent, title=title, style=style)

        fslgl.getGLContext()
        fslgl.bootstrap()

        self.panel = OrthoPanel(self, overlayList, displayCtx, None)
        self.Layout()



#!/usr/bin/env python
#
# __init__.py - Funtions for managing OpenGL shader programs.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""The ``shaders`` package contains classes and functions for finding,
parsing, compiling, and managing OpenGL shader programs. Two types of shader
program are supported:


 - GLSL 1.20 vertex and fragment shaders.

 - ``ARB_vertex_program`` and ``ARB_fragment_program`` shader programs.


The :mod:`.glsl` and :mod:`.arbp` packages respectively define the
:class:`.GLSLShader` and :class:`.ARBPShader` classes, which may be
used to manage shader programs of the corresponding type.


Some package-level functions are defined here, for finding and loading
shader source code:


  .. autosummary::
     :nosignatures:

     getShaderDir
     getShaderSuffix
     getVertexShader
     getFragmentShader
"""


# Import open from the io module, because it gives
# us an interface compatible across python 2 and 3
# (i.e. it allows us to specify the file encoding,
# and thus allows shader files to contain non-ascii
# characters).
from io import                 open
import os.path              as op

import                         fsleyes
import fsleyes.gl           as fslgl
from   .glsl import program as glslprogram
from   .arbp import program as arbpprogram


GLSLShader = glslprogram.GLSLShader
ARBPShader = arbpprogram.ARBPShader


def getShaderDir():
    """Returns the directory in which the ``ARB`` and ``glsl`` shader program
    source files can be found. A different directory will be returned depending
    on which OpenGL version is in use.
    """

    if   fslgl.GL_VERSION == '2.1': subdir = 'gl21'
    elif fslgl.GL_VERSION == '1.4': subdir = 'gl14'

    return op.join(fsleyes.assetDir, 'assets', 'gl', subdir)


def getShaderSuffix():
    """Returns the shader program file suffix to use. A different suffix will be
    returned depending on which OpenGL version is in use.
    """

    if   fslgl.GL_VERSION == '2.1': return 'glsl'
    elif fslgl.GL_VERSION == '1.4': return 'prog'


def getVertexShader(prefix):
    """Returns the vertex shader source for the given GL type (e.g.
    'glvolume').
    """
    return _getShader(prefix, 'vert')


def getFragmentShader(prefix):
    """Returns the fragment shader source for the given GL type."""
    return _getShader(prefix, 'frag')


def _getShader(prefix, shaderType):
    """Returns the shader source for the given GL type and the given
    shader type ('vert' or 'frag').
    """
    fname = _getFileName(prefix, shaderType)
    with open(fname, 'rt', encoding='utf-8') as f:
        src = f.read()

    return preprocess(src)


def _getFileName(prefix, shaderType):
    """Returns the file name of the shader program for the given GL type
    and shader type.
    """

    suffix = getShaderSuffix()

    if shaderType not in ('vert', 'frag'):
        raise RuntimeError('Invalid shader type: {}'.format(shaderType))

    return op.join(getShaderDir(), '{}_{}.{}'.format(
        prefix, shaderType, suffix))


def preprocess(src):
    """'Preprocess' the given shader source.

    This amounts to searching for lines containing '#pragma include filename',
    and replacing those lines with the contents of the specified files.
    """

    lines    = src.split('\n')
    lines    = [l.strip() for l in lines]

    pragmas = []
    for linei, line in enumerate(lines):
        if line.startswith('#pragma'):
            pragmas.append(linei)

    includes = []
    for linei in pragmas:

        line = lines[linei].split()

        if len(line) != 3:       continue
        if line[1] != 'include': continue

        includes.append((linei, line[2]))

    for linei, fname in includes:
        fname = op.join(getShaderDir(), fname)
        with open(fname, 'rt', encoding='utf-8') as f:
            lines[linei] = f.read()

    return '\n'.join(lines)



#!/usr/bin/env python
#
# texture3d.py - The Texture3D class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`Texture3D` class, which represents a
3D OpenGL texture.
"""


import logging

import numpy          as np
import OpenGL.GL      as gl

import fsl.data.utils as dutils

from . import            texture


log = logging.getLogger(__name__)


class Texture3D(texture.Texture):
    """The ``Texture3D`` class contains the logic required to create and
    manage a 3D texture.
    """


    def __init__(self, name, **kwargs):
        """Create a ``Texture3D``.

        :arg name: A unique name for the texture.

        All other keyword arguments are passed through to
        :meth:`.Texture.__init__`.
        """

        kwargs['nvals'] = kwargs.get('nvals', 1)

        if kwargs['nvals'] not in (1, 3, 4):
            raise ValueError('nvals must be either 1, 3 or 4')

        texture.Texture.__init__(self, name, 3, **kwargs)


    def doRefresh(self):
        """Overrides :meth:`.Texture.doRefresh`.

        (Re-)configures the OpenGL texture.
        """

        data = self.preparedData

        if data is None:
            return

        log.debug('Configuring 3D texture (id %s) for %s (data shape: %s)',
                  self.name, self.handle, self.shape)

        # First dimension for multi-
        # valued textures
        if self.nvals > 1: shape = data.shape[1:]
        else:              shape = data.shape

        # The image data is flattened, with
        # fortran dimension ordering, so the
        # data, as stored on the GPU, has its
        # first dimension as the fastest
        # changing.
        data = np.array(data.ravel(order='F'), copy=False)

        # PyOpenGL needs the data array
        # to be writeable, as it uses
        # PyArray_ISCARRAY to check
        # for contiguousness. but if the
        # data has come from a nibabel
        # ArrayProxy, the writeable flag
        # will be set to False for some
        # reason.
        data    = dutils.makeWriteable(data)

        interp  = self.interp
        intFmt  = self.internalFormat
        baseFmt = self.baseFormat
        ttype   = self.textureType

        if interp is None:
            interp = gl.GL_NEAREST

        with self.bound():

            # Enable storage of tightly packed data of any size (i.e.
            # our texture shape does not have to be divisible by 4).
            gl.glPixelStorei(gl.GL_PACK_ALIGNMENT,   1)
            gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, 1)

            gl.glTexParameteri(gl.GL_TEXTURE_3D,
                               gl.GL_TEXTURE_MAG_FILTER,
                               interp)
            gl.glTexParameteri(gl.GL_TEXTURE_3D,
                               gl.GL_TEXTURE_MIN_FILTER,
                               interp)

            # Clamp texture borders to the edge
            # values - it is the responsibility
            # of the rendering logic to not draw
            # anything outside of the image space
            gl.glTexParameteri(gl.GL_TEXTURE_3D,
                               gl.GL_TEXTURE_WRAP_S,
                               gl.GL_CLAMP_TO_EDGE)
            gl.glTexParameteri(gl.GL_TEXTURE_3D,
                               gl.GL_TEXTURE_WRAP_T,
                               gl.GL_CLAMP_TO_EDGE)
            gl.glTexParameteri(gl.GL_TEXTURE_3D,
                               gl.GL_TEXTURE_WRAP_R,
                               gl.GL_CLAMP_TO_EDGE)

            # The macOS GL driver sometimes corrupts
            # the texture data if we don't generate
            # mipmaps
            gl.glTexParameteri(gl.GL_TEXTURE_3D,
                               gl.GL_GENERATE_MIPMAP,
                               gl.GL_TRUE)

            # create the texture according to
            # the format determined by the
            # determineTextureType method.
            #
            # note: The ancient Chromium driver (still
            #       in use by VirtualBox) will improperly
            #       create 3D textures without two calls
            #       (to glTexImage3D and glTexSubImage3D).
            #       If I specify the texture size and set
            #       the data in a single call, it seems to
            #       expect that the data or texture
            #       dimensions always have even size - odd
            #       sized images will be displayed
            #       incorrectly.
            gl.glTexImage3D(gl.GL_TEXTURE_3D,
                            0,
                            intFmt,
                            shape[0],
                            shape[1],
                            shape[2],
                            0,
                            baseFmt,
                            ttype,
                            None)
            gl.glTexSubImage3D(gl.GL_TEXTURE_3D,
                               0, 0, 0, 0,
                               shape[0],
                               shape[1],
                               shape[2],
                               baseFmt,
                               ttype,
                               data)


    def doPatch(self, data, offset):
        """Overrides :meth:`.Texture.doPatch`. Updates part of the texture
        data.
        """

        shape = data.shape
        data  = data.flatten(order='F')

        with self.bound():
            gl.glTexSubImage3D(gl.GL_TEXTURE_3D,
                               0,
                               offset[0],
                               offset[1],
                               offset[2],
                               shape[0],
                               shape[1],
                               shape[2],
                               self.baseFormat,
                               self.textureType,
                               data)



#!/usr/bin/env python
#
# mipopts.py - The MIPOpts class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`MIPOpts` class, a :class:`.DisplayOpts`
class containing options for rendering maximum intensity projections of
:class:`.Image` overlays. The :class:`MIPOpts` class corresponds to the
``'mip'`` overlay type.

MIP overlays are rendered via the :class:`.GLMIP` class.
"""


import numpy as np

from   fsl.utils.platform import platform as fslplatform
import fsl.transform.affine               as affine
import fsleyes_props                      as props

from . import colourmapopts               as cmapopts
from . import                                niftiopts


class MIPOpts(cmapopts.ColourMapOpts, niftiopts.NiftiOpts):
    """The ``MIPOpts`` class is used for rendering maximum intensity
    projections of .Image overlays.
    """


    window = props.Percentage(minval=1, clamped=True, default=50)
    """Window over which the MIP is calculated, as a proportion of the image
    length. The window is centered at the current display location.
    """


    minimum = props.Boolean(default=False)
    """Display a minimum intensity projection, rather than maximum. """


    absolute = props.Boolean(default=False)
    """Display an absolute maximum intensity projection. This setting
    overrides the :attr:`minimum` setting.
    """


    interpolation = props.Choice(('none', 'linear', 'spline'))
    """How the value shown at a real world location is derived from the
    corresponding data value(s). ``none`` is equivalent to nearest neighbour
    interpolation.
    """


    def __init__(self, *args, **kwargs):
        """Create a ``MIPOpts`` object.

        All arguments are passed through to the :class:`.NiftiOpts` init
        function.
        """

        # We need GL >= 2.1 for
        # spline interpolation
        if float(fslplatform.glVersion) < 2.1:
            interp = self.getProp('interpolation')
            interp.removeChoice('spline', instance=self)
            interp.updateChoice('linear', instance=self, newAlt=['spline'])

        niftiopts.NiftiOpts    .__init__(self, *args, **kwargs)
        cmapopts .ColourMapOpts.__init__(self)

        # calculate the approximate number
        # of voxels along the longest diagonal
        # of the image - we use this to calculate
        # the maximum number of samples to take
        x, y, z  = self.overlay.shape[:3]
        xy       = (x * y, (x, y))
        xz       = (x * z, (x, z))
        yz       = (y * z, (y, z))
        ax0, ax1 = max((xy, xz, yz))[1]
        self.numSteps = np.ceil(np.sqrt(ax0 ** 2 + ax1 ** 2)) * 2


    def destroy(self):
        """Must be called when this ``MIPOpts`` object is no longer needed. """
        cmapopts .ColourMapOpts.destroy(self)
        niftiopts.NiftiOpts    .destroy(self)


    def getDataRange(self):
        """Overrides :meth:`.ColourMapOpts.getDataRange`. Returns the
        :attr:`.Image.dataRange` of the image.
        """
        return self.overlay.dataRange


    def calculateRayCastSettings(self, viewmat):
        """Calculates a camera direction and ray casting step vector, based
        on the given view matrix.
        """

        d2tmat  = self.getTransform('display', 'texture')
        xform   = affine.concat(d2tmat, viewmat)
        cdir    = np.array([0, 0, 1])
        cdir    = affine.transform(cdir, xform, vector=True)
        cdir    = affine.normalise(cdir)

        # sqrt(3) so the maximum number
        # of samplews is taken along the
        # diagonal of a cube
        rayStep = np.sqrt(3) * cdir / self.numSteps

        return cdir, rayStep



#!/usr/bin/env python
#
# display.py - Definitions of the Display and DisplayOpts classes.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`Display` and :class:`DisplayOpts` classes,
which encapsulate overlay display settings.
"""


import logging
import inspect

import fsl.data.image                 as fslimage
import fsl.data.constants             as constants
import fsl.utils.deprecated           as deprecated
import fsleyes_props                  as props
import fsleyes_widgets.utils.typedict as td

import fsleyes.strings                as strings
import fsleyes.actions                as actions


log = logging.getLogger(__name__)


class Display(props.SyncableHasProperties):
    """The ``Display`` class contains display settings which are common to
    all overlay types.

    A ``Display`` instance is also responsible for managing a single
    :class:`DisplayOpts` instance, which contains overlay type specific
    display options. Whenever the :attr:`overlayType` property of a
    ``Display`` instance changes, the old ``DisplayOpts`` instance (if any)
    is destroyed, and a new one, of the correct type, created.
    """


    name = props.String()
    """The overlay name. """


    overlayType = props.Choice()
    """This property defines the overlay type - how the data is to be
    displayed.

    The options for this property are populated in the :meth:`__init__`
    method, from the :attr:`.displaycontext.OVERLAY_TYPES` dictionary. A
    :class:`DisplayOpts` sub-class exists for every possible value that this
    property may take.

    """

    enabled = props.Boolean(default=True)
    """Should this overlay be displayed at all? """


    alpha = props.Percentage(default=100.0)
    """Opacity - 100% is fully opaque, and 0% is fully transparent."""


    brightness = props.Percentage()
    """Brightness - 50% is normal brightness."""


    contrast   = props.Percentage()
    """Contrast - 50% is normal contrast."""


    def __init__(self,
                 overlay,
                 overlayList,
                 displayCtx,
                 parent=None,
                 **kwa):
        """Create a :class:`Display` for the specified overlay.

        :arg overlay:     The overlay object.

        :arg overlayList: The :class:`.OverlayList` instance which contains
                          all overlays.

        :arg displayCtx:  A :class:`.DisplayContext` instance describing how
                          the overlays are to be displayed.

        :arg parent:      A parent ``Display`` instance - see
                          :mod:`props.syncable`.

        All other keyword arguments are assumed to be ``name=value`` pairs,
        containing initial property values, for both this ``Display``, and
        the initially created :class:`DisplayOpts` instance. For the latter,
        it is assumed that any properties specified are appropriate for the
        initial overlay type.
        """

        dispProps     = self.getAllProperties()[0]
        initDispProps = {n : v for n, v in kwa.items() if n     in dispProps}
        initOptProps  = {n : v for n, v in kwa.items() if n not in dispProps}

        self.__overlay     = overlay
        self.__overlayList = overlayList
        self.__displayCtx  = displayCtx
        self.name          = overlay.name

        # Populate the possible choices
        # for the overlayType property
        from . import getOverlayTypes

        ovlTypes    = getOverlayTypes(overlay)
        ovlTypeProp = self.getProp('overlayType')

        log.debug('Enabling overlay types for {}: '.format(overlay, ovlTypes))
        ovlTypeProp.setChoices(ovlTypes, instance=self)

        # Call the super constructor after our own
        # initialisation, in case the provided parent
        # has different property values to our own,
        # and our values need to be updated
        props.SyncableHasProperties.__init__(
            self,
            parent=parent,

            # These properties cannot be unbound, as
            # they affect the OpenGL representation.
            # The name can't be unbound either,
            # because it would be silly to allow
            # different names for the same overlay.
            nounbind=['overlayType', 'name'],

            # Initial sync state between this
            # Display and the parent Display
            # (if this Display has a parent)
            state=displayCtx.syncOverlayDisplay,

            # set initial display property values
            **initDispProps)

        # When the overlay type changes, the property
        # values of the DisplayOpts instance for the
        # old overlay type are stored in this dict.
        # If the overlay is later changed back to the
        # old type, its previous values are restored.
        #
        # The structure of the dictionary is:
        #
        #   { (type(DisplayOpts), propName) : propValue }
        #
        # This also applies to the case where the
        # overlay type is changed from one type to
        # a related type (e.g. from VolumeOpts to
        # LabelOpts) - the values of all common
        # properties are copied to the new
        # DisplayOpts instance.
        self.__oldOptProps = td.TypeDict()

        # Initial DisplayOpt property values
        # are used in the first call to
        # __makeDisplayOpts, and then cleared
        # afterwards.
        self.__initOptProps = initOptProps

        # Set up listeners after caling Syncable.__init__,
        # so the callbacks don't get called during
        # synchronisation
        self.addListener(
            'overlayType',
            'Display_{}'.format(id(self)),
            self.__overlayTypeChanged)

        # The __overlayTypeChanged method creates
        # a new DisplayOpts instance - for this,
        # it needs to be able to access this
        # Display instance's parent (so it can
        # subsequently access a parent for the
        # new DisplayOpts instance). Therefore,
        # we do this after calling Syncable.__init__.
        self.__displayOpts = None
        self.__overlayTypeChanged()

        log.debug('{}.init ({})'.format(type(self).__name__, id(self)))


    def __del__(self):
        """Prints a log message."""
        if log:
            log.debug('{}.del ({})'.format(type(self).__name__, id(self)))


    def destroy(self):
        """This method must be called when this ``Display`` instance
        is no longer needed.

        When a ``Display`` instance is destroyed, the corresponding
        :class:`DisplayOpts` instance is also destroyed.
        """

        if self.__displayOpts is not None:
            self.__displayOpts.destroy()

        self.removeListener('overlayType', 'Display_{}'.format(id(self)))

        self.detachAllFromParent()

        self.__oldOptProps  = None
        self.__initOptProps = None
        self.__displayOpts  = None
        self.__overlayList  = None
        self.__displayCtx   = None
        self.__overlay      = None


    @deprecated.deprecated('0.14.3', '1.0.0', 'Use overlay instead')
    def getOverlay(self):
        """Deprecated - use :meth:`overlay` instead."""
        return self.__overlay


    @property
    def overlay(self):
        """Returns the overlay associated with this ``Display`` instance."""
        return self.__overlay


    @property
    def opts(self):
        """Return the :class:`.DisplayOpts` instance associated with this
        ``Display``, which contains overlay type specific display settings.

        If a ``DisplayOpts`` instance has not yet been created, or the
        :attr:`overlayType` property no longer matches the type of the
        existing ``DisplayOpts`` instance, a new ``DisplayOpts`` instance
        is created (and the old one destroyed if necessary).

        See the :meth:`__makeDisplayOpts` method.
        """

        if (self.__displayOpts             is None) or \
           (self.__displayOpts.overlayType != self.overlayType):

            if self.__displayOpts is not None:
                self.__displayOpts.destroy()

            self.__displayOpts = self.__makeDisplayOpts()

        return self.__displayOpts


    @deprecated.deprecated('0.16.0', '1.0.0', 'Use opts instead')
    def getDisplayOpts(self):
        """Return the :class:`.DisplayOpts` instance associated with this
        ``Display``, which contains overlay type specific display settings.

        If a ``DisplayOpts`` instance has not yet been created, or the
        :attr:`overlayType` property no longer matches the type of the
        existing ``DisplayOpts`` instance, a new ``DisplayOpts`` instance
        is created (and the old one destroyed if necessary).

        See the :meth:`__makeDisplayOpts` method.
        """

        if (self.__displayOpts             is None) or \
           (self.__displayOpts.overlayType != self.overlayType):

            if self.__displayOpts is not None:
                self.__displayOpts.destroy()

            self.__displayOpts = self.__makeDisplayOpts()

        return self.__displayOpts


    def __makeDisplayOpts(self):
        """Creates a new :class:`DisplayOpts` instance. The specific
        ``DisplayOpts`` sub-class that is created is dictated by the current
        value of the :attr:`overlayType` property.

        The :data:`.displaycontext.DISPLAY_OPTS_MAP` dictionary defines the
        mapping between overlay types and :attr:`overlayType` values, and
        ``DisplayOpts`` sub-class types.
        """

        if self.getParent() is None:
            oParent = None
        else:
            oParent = self.getParent().opts

        initOptProps        = self.__initOptProps
        self.__initOptProps = None

        if initOptProps is None:
            initOptProps = {}

        from . import DISPLAY_OPTS_MAP

        optType = DISPLAY_OPTS_MAP[self.__overlay, self.overlayType]

        log.debug('Creating {} instance (synced: {}) for overlay '
                  '{} ({})'.format(optType.__name__,
                                   self.__displayCtx.syncOverlayDisplay,
                                   self.__overlay, self.overlayType))

        volProps  = optType.getVolumeProps()
        allProps  = optType.getAllProperties()[0]
        initState = {}

        for p in allProps:
            if p in volProps:
                initState[p] = self.__displayCtx.syncOverlayVolume
            else:
                initState[p] = self.__displayCtx.syncOverlayDisplay

        return optType(self.__overlay,
                       self,
                       self.__overlayList,
                       self.__displayCtx,
                       parent=oParent,
                       state=initState,
                       **initOptProps)


    def __findOptBaseType(self, optType, optName):
        """Finds the class, in the hierarchy of the given ``optType`` (a
        :class:`.DisplayOpts` sub-class) in which the given ``optName``
        is defined.

        This method is used by the :meth:`__saveOldDisplayOpts` method, and
        is an annoying necessity caused by the way that the :class:`.TypeDict`
        class works. A ``TypeDict`` does not allow types to be used as keys -
        they must be strings containing the type names.

        Furthermore, in order for the property values of a common
        ``DisplayOpts`` base type to be shared across sub types (e.g. copying
        the :attr:`.NiftiOpts.transform` property between :class:`.VolumeOpts`
        and :class:`.LabelOpts` instances), we need to store the name of the
        common base type in the dictionary.
        """

        for base in inspect.getmro(optType):
            if optName in base.__dict__:
                return base

        return None


    def __saveOldDisplayOpts(self):
        """Saves the value of every property on the current
        :class:`DisplayOpts` instance, so they can be restored later if
        needed.
        """

        opts = self.__displayOpts

        if opts is None:
            return

        for propName in opts.getAllProperties()[0]:
            base = self.__findOptBaseType(type(opts), propName)
            base = base.__name__
            val  = getattr(opts, propName)

            log.debug('Saving {}.{} = {} [{} {}]'.format(
                base, propName, val, type(opts).__name__, id(self)))

            self.__oldOptProps[base, propName] = val


    def __restoreOldDisplayOpts(self):
        """Restores any cached values for all of the properties on the
        current :class:`DisplayOpts` instance.
        """
        opts = self.__displayOpts

        if opts is None:
            return

        for propName in opts.getAllProperties()[0]:

            try:
                value = self.__oldOptProps[opts, propName]

                if not hasattr(opts, propName):
                    continue

                if not opts.propertyIsEnabled(propName):
                    continue

                log.debug('Restoring {}.{} = {} [{}]'.format(
                    type(opts).__name__, propName, value, id(self)))

                setattr(opts, propName, value)

            except KeyError:
                pass


    def __overlayTypeChanged(self, *a):
        """Called when the :attr:`overlayType` property changes. Makes sure
        that the :class:`DisplayOpts` instance is of the correct type.
        """
        self.__saveOldDisplayOpts()
        self.opts
        self.__restoreOldDisplayOpts()


class DisplayOpts(props.SyncableHasProperties, actions.ActionProvider):
    """The ``DisplayOpts`` class contains overlay type specific display
    settings. ``DisplayOpts`` instances are managed by :class:`Display`
    instances.


    The ``DisplayOpts`` class is not meant to be created directly - it is a
    base class for type specific implementations (e.g. the :class:`.VolumeOpts`
    class).


    The following attributes are available on all ``DisplayOpts`` instances:


    =============== ======================================================
    ``overlay``     The overlay object
    ``display``     The :class:`Display` instance that created this
                    ``DisplayOpts`` instance.
    ``overlayType`` The value of the :attr:`Display.overlayType` property
                    corresponding to the type of this ``DisplayOpts``
                    instance.
    ``overlayList`` The :class:`.OverlayList` instance, which contains all
                    overlays.
    ``displayCtx``  The :class:`.DisplayContext` instance which is
                    responsible for all ``Display`` and ``DisplayOpts``
                    instances.
    ``name``        A unique name for this ``DisplayOpts`` instance.
    =============== ======================================================


    .. warning:: :class:`DisplayOpts` sub-classes must not define any
                 properties with the same name as any of the :class:`Display`
                 properties.
    """


    bounds = props.Bounds(ndims=3)
    """Specifies a bounding box in the display coordinate system which is big
    enough to contain the overlay described by this ``DisplayOpts``
    instance.

    The values in this ``bounds`` property must be updated by ``DisplayOpts``
    subclasses whenever the spatial representation of their overlay changes.
    """


    def __init__(
            self,
            overlay,
            display,
            overlayList,
            displayCtx,
            **kwargs):
        """Create a ``DisplayOpts`` object.

        :arg overlay:     The overlay associated with this ``DisplayOpts``
                          instance.

        :arg display:     The :class:`Display` instance which owns this
                          ``DisplayOpts`` instance.

        :arg overlayList: The :class:`.OverlayList` which contains all
                          overlays.

        :arg displayCtx:  A :class:`.DisplayContext` instance describing
                          how the overlays are to be displayed.
        """

        self.__overlay     = overlay
        self.__display     = display
        self.__overlayType = display.overlayType
        self.__name        = '{}_{}'.format(type(self).__name__, id(self))

        props.SyncableHasProperties.__init__(self, **kwargs)
        actions.ActionProvider     .__init__(self, overlayList, displayCtx)

        log.debug('{}.init [DC: {}] ({})'.format(
            type(self).__name__, id(displayCtx), id(self)))


    def __del__(self):
        """Prints a log message."""
        if log:
            log.debug('{}.del ({})'.format(type(self).__name__, id(self)))


    def destroy(self):
        """This method must be called when this ``DisplayOpts`` instance
        is no longer needed.

        If a subclass overrides this method, the subclass implementation
        must call this method, **after** performing its own clean up.
        """
        actions.ActionProvider.destroy(self)

        self.detachAllFromParent()

        self.__overlay = None
        self.__display = None


    @classmethod
    def getVolumeProps(cls):
        """Intended to be overridden by sub-classes as needed.  Returns a list
        of property names which control the currently displayed
        volume/timepoint for 4D overlays. The default implementation returns
        an empty list.
        """
        return []


    @property
    def overlay(self):
        """Return the overlay associated with this ``DisplayOpts`` object.
        """
        return self.__overlay


    @property
    def display(self):
        """Return the :class:`.Display` that is managing this
        ``DisplayOpts`` object.
        """
        return self.__display


    @property
    def overlayType(self):
        """Return the type of this ``DisplayOpts`` object (the value of
        :attr:`Display.overlayType`).
        """
        return self.__overlayType


    @property
    def name(self):
        """Return the name of this ``DisplayOpts`` object. """
        return self.__name


    @property
    def referenceImage(self):
        """Return the reference image associated with this ``DisplayOpts``
        instance.

        Some non-volumetric overlay types (e.g. the :class:`.Mesh` -
        see :class:`.MeshOpts`) may have a *reference* :class:`.Nifti` instance
        associated with them, allowing the overlay to be localised in the
        coordinate space defined by the :class:`.Nifti`. The
        :class:`.DisplayOpts` sub-class which corresponds to
        such non-volumetric overlays should override this method to return
        that reference image.

        :class:`.DisplayOpts` sub-classes which are associated with volumetric
        overlays (i.e. :class:`.Nifti` instances) do not need to override
        this method - in this case, the overlay itself is considered to be
        its own reference image, and is returned by the base-class
        implementation of this method.

        .. note:: The reference :class:`.Nifti` instance returned by
                  sub-class implementations of this method must be in
                  the :class:`.OverlayList`.
        """

        if isinstance(self.overlay, fslimage.Nifti):
            return self.overlay
        return None


    @deprecated.deprecated('0.14.3', '1.0.0', 'Use referenceImage instead')
    def getReferenceImage(self):
        """Deprecated - use :meth:`referenceImage` instead. """
        return self.referenceImage


    def getLabels(self):
        """Generates some orientation labels for the overlay associated with
        this ``DisplayOpts`` instance.

        If the overlay is not a ``Nifti`` instance, or does not have a
        reference image set, the labels will represent an unknown orientation.

        Returns a tuple containing:

          - The ``(xlo, ylo, zlo, xhi, yhi, zhi)`` labels
          - The ``(xorient, yorient, zorient)`` orientations (see
            :meth:`.Image.getOrientation`)
        """

        refImage = self.referenceImage

        if refImage is None:
            return ('??????', [constants.ORIENT_UNKNOWN] * 3)

        opts = self.displayCtx.getOpts(refImage)

        xorient = None
        yorient = None
        zorient = None

        # If we are displaying in voxels/scaled voxels,
        # and this image is not the current display
        # image, then we do not show anatomical
        # orientation labels, as there's no guarantee
        # that all of the loaded overlays are in the
        # same orientation, and it can get confusing.
        if opts.transform in ('id', 'pixdim', 'pixdim-flip') and \
           self.displayCtx.displaySpace != refImage:
            xlo = 'Xmin'
            xhi = 'Xmax'
            ylo = 'Ymin'
            yhi = 'Ymax'
            zlo = 'Zmin'
            zhi = 'Zmax'

        # Otherwise we assume that all images
        # are aligned to each other, so we
        # estimate the current image's orientation
        # in the display coordinate system
        else:

            xform   = opts.getTransform('display', 'world')
            xorient = refImage.getOrientation(0, xform)
            yorient = refImage.getOrientation(1, xform)
            zorient = refImage.getOrientation(2, xform)

            xlo     = strings.anatomy['Nifti', 'lowshort',  xorient]
            ylo     = strings.anatomy['Nifti', 'lowshort',  yorient]
            zlo     = strings.anatomy['Nifti', 'lowshort',  zorient]
            xhi     = strings.anatomy['Nifti', 'highshort', xorient]
            yhi     = strings.anatomy['Nifti', 'highshort', yorient]
            zhi     = strings.anatomy['Nifti', 'highshort', zorient]

        return ((xlo, ylo, zlo, xhi, yhi, zhi),
                (xorient, yorient, zorient))



#!/usr/bin/env python
#
# exportdataseries.py - The ExportDataSeriesAction.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ExportDataSeriesAction`, which is used by
:class:`.PlotPanel` views to export data series to a text file.
"""


import                          os
import os.path               as op
import numpy                 as np

import fsl.utils.settings    as fslsettings
import fsleyes.strings       as strings
from . import                   base


class ExportDataSeriesAction(base.Action):

    def __init__(self, overlayList, displayCtx, plotPanel):

        base.Action.__init__(self, overlayList, displayCtx, self.__doExport)
        self.__plotPanel = plotPanel


    def __doExport(self):

        import wx

        # Ask the user if they want to save the
        # x axis values as the first column
        dlg = wx.MessageDialog(
            wx.GetTopLevelWindows()[0],
            message=strings.messages[self, 'saveXColumn'],
            caption=strings.titles[  self, 'saveXColumn'],
            style=(wx.ICON_QUESTION |
                   wx.YES_NO        |
                   wx.CANCEL        |
                   wx.YES_DEFAULT))

        result = dlg.ShowModal()

        if result == wx.ID_CANCEL:
            return

        savex = result == wx.ID_YES

        # Ask the user where they want to save the data
        msg     = strings.messages[self, 'selectFile']
        fromDir = fslsettings.read('loadSaveOverlayDir', os.getcwd())
        dlg     = wx.FileDialog(wx.GetApp().GetTopWindow(),
                                message=msg,
                                defaultDir=fromDir,
                                defaultFile='dataseries.txt',
                                style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)

        if dlg.ShowModal() != wx.ID_OK:
            return

        filePath = dlg.GetPath()

        dss = self.__plotPanel.getDrawnDataSeries()
        xs  = [ds[1] for ds in dss]
        ys  = [ds[2] for ds in dss]
        dss = [ds[0] for ds in dss]

        # Create some x data, unified
        # across all data series
        xdata = np.unique(np.concatenate(xs))

        # Linearly interpolate each data series
        # according to the merged x data
        ydata = [np.interp(xdata, x, y, left=np.nan, right=np.nan)
                 for x, y in zip(xs, ys)]

        # Turn it all into one big
        # array and save it out
        if savex: data = np.vstack([xdata] + ydata)
        else:     data = np.vstack( ydata)

        np.savetxt(filePath, data.T, fmt='% 0.8f')

        fslsettings.write('loadSaveOverlayDir', op.dirname(filePath))



#!/usr/bin/env python
#
# overlayinfopanel.py - The OverlayInfoPanel class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the ``OverlayInfoPanel`` class, a *FSLeyes control*
panel which displays information about the currently selected overlay.
"""


import logging
import collections

import wx

import numpy as np

import fsl.utils.bids                 as fslbids
import fsl.data.image                 as fslimage
import fsl.data.constants             as constants
import fsl.transform.affine           as affine
import fsleyes_widgets.utils.typedict as td

import fsleyes.controls.controlpanel  as ctrlpanel
import fsleyes.strings                as strings


USE_HTML2 = False
"""Toggle this flag to switch between the simple wx.html renderer,
and the webkit-backed wx.html2 renderer. Webkit is not necessarily
present on all systems, and there's no neat way to dynamically
test whether wx.html2 will work. So I'm sticking with wx.html for
now.
"""


if USE_HTML2: import wx.html2 as wxhtml
else:         import wx.html  as wxhtml


log = logging.getLogger(__name__)


# The wx.html2.WebView.SetPage method differs from
# the wx.html.HtmlWindow.SetPage method - it requires
# two parameters. Here we're sub-classing the
# HtmlWindow method so that it also accepts two
# parameters, but ignores the second.
#
# The sub-class also forces links to be opened
# in the system browser, as opposed to in the
# same window.
if not USE_HTML2:

    class HtmlWindow(wxhtml.HtmlWindow):

        def __init__(self, *args, **kwargs):
            wxhtml.HtmlWindow.__init__(self, *args, **kwargs)

            self.Bind(wxhtml.EVT_HTML_LINK_CLICKED, self.__onLink)

            # wx.html.HtmlWindow defaults
            # to a slightly bigger font size
            self.SetStandardFonts(self.GetFont().GetPointSize())

        def __onLink(self, ev):
            wx.LaunchDefaultBrowser(ev.GetLinkInfo().GetHref())

        def SetPage(self, html, url=None):
            wxhtml.HtmlWindow.SetPage(self, html)


class OverlayInfoPanel(ctrlpanel.ControlPanel):
    """An ``OverlayInfoPanel`` is a :class:`.ControlPanel` which displays
    information about the currently selected overlay in a
    ``wx.html.HtmlWindow``. The currently selected overlay is defined by the
    :attr:`.DisplayContext.selectedOverlay` property. An ``OverlayInfoPanel``
    looks something like the following:

    .. image:: images/overlayinfopanel.png
       :scale: 50%
       :align: center

    Slightly different information is shown depending on the overlay type,
    and is generated by the following methods:

    ======================== =============================
    :class:`.Image`          :meth:`__getImageInfo`
    :class:`.FEATImage`      :meth:`__getFEATImageInfo`
    :class:`.MelodicImage`   :meth:`__getMelodicImageInfo`
    :class:`.DTIFitTensor`   :meth:`__getDTIFitTensorInfo`
    :class:`.Mesh`           :meth:`__getMeshInfo`
    :class:`.VTKMesh`        :meth:`__getVTKMeshInfo`
    :class:`.GiftiMesh`      :meth:`__getGiftiMeshInfo`
    :class:`.FreesurferMesh` :meth:`__getFreesurferMeshInfo`
    ======================== =============================
    """


    def __init__(self, parent, overlayList, displayCtx, frame):
        """Create an ``OverlayInfoPanel``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg frame:       The :class:`.FSLeyesFrame` instance.
        """

        ctrlpanel.ControlPanel.__init__(
            self, parent, overlayList, displayCtx, frame)

        if USE_HTML2: self.__info = wxhtml.WebView.New(self)
        else:         self.__info = HtmlWindow(        self)

        self.__sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.__sizer.Add(self.__info, flag=wx.EXPAND, proportion=1)

        self.SetSizer(self.__sizer)

        displayCtx .addListener('selectedOverlay',
                                self.name,
                                self.__selectedOverlayChanged)
        overlayList.addListener('overlays',
                                self.name,
                                self.__selectedOverlayChanged)

        self.__currentOverlay = None
        self.__currentDisplay = None
        self.__currentOpts    = None
        self.__selectedOverlayChanged()

        self.SetMinSize((350, 500))
        self.Layout()


    def destroy(self):
        """Must be called when this ``OverlayInfoPanel`` is no longer
        needed. Removes some property listeners, and calls the
        :meth:`.ControlPanel.destroy` method.
        """

        self.displayCtx .removeListener('selectedOverlay', self.name)
        self.overlayList.removeListener('overlays',        self.name)

        self.__deregisterOverlay()

        ctrlpanel.ControlPanel.destroy(self)


    def __selectedOverlayChanged(self, *a):
        """Called when the :class:`.OverlayList` or
        :attr:`.DisplayContext.selectedOverlay` changes. Refreshes the
        information shown on this ``OverlayInfoPanel``.
        """

        overlay = self.displayCtx.getSelectedOverlay()

        # Overlay list is empty
        if overlay is None:
            self.__info.SetPage('', '')
            self.__info.Refresh()
            return

        self.__deregisterOverlay()

        if overlay is not None:
            self.__registerOverlay(overlay)

        self.__updateInformation()


    _optProps = td.TypeDict({
        'Image'        : ['transform'],
        'Mesh'         : ['refImage', 'coordSpace'],
        'DTIFitTensor' : ['transform'],
    })
    """This dictionary contains a list of :class:`.DisplayOpts` properties
    that, when changed, should result in the information being refreshed.
    It is used by the :meth:`__registerOverlay` and :meth:`__deregisterOverlay`
    methods.
    """


    def __registerOverlay(self, overlay):
        """Registers property listeners with the given overlay so the
        information can be refreshed when necessary.
        """

        display = self.displayCtx.getDisplay(overlay)
        opts    = display.opts

        self.__currentOverlay = overlay
        self.__currentDisplay = display
        self.__currentOpts    = opts

        display.addListener('name',
                            self.name,
                            self.__overlayNameChanged)
        display.addListener('overlayType',
                            self.name,
                            self.__overlayTypeChanged)

        for propName in OverlayInfoPanel._optProps.get(overlay, []):
            opts.addListener(propName, self.name, self.__overlayOptsChanged)


    def __deregisterOverlay(self):
        """De-registers property listeners from the overlay that was
        previously registered via :meth:`__registerOverlay`.
        """

        if self.__currentOverlay is None:
            return

        overlay = self.__currentOverlay
        display = self.__currentDisplay
        opts    = self.__currentOpts

        self.__currentOverlay = None
        self.__currentDisplay = None
        self.__currentOpts    = None

        display.removeListener('name',        self.name)
        display.removeListener('overlayType', self.name)

        for propName in OverlayInfoPanel._optProps[overlay]:
            opts.removeListener(propName, self.name)


    def __overlayTypeChanged(self, *a):
        """Called when the :attr:`.Display.overlayType` for the current
        overlay changes. Re-registers with the ``Display`` and
        ``DisplayOpts`` instances associated with the overlay.
        """
        self.__selectedOverlayChanged()


    def __overlayNameChanged(self, *a):
        """Called when the :attr:`.Display.name` for the current overlay
        changes. Updates the information display.
        """
        self.__updateInformation()


    def __overlayOptsChanged(self, *a):
        """Called when any :class:`.DisplayOpts` properties for the current
        overlay change. Updates the information display. The properties that
        trigger a refresh are  defined in the :attr:`_optProps` dictionary.
        """
        self.__updateInformation()


    def __updateInformation(self):
        """Refreshes the information shown on this ``OverlayInfoPanel``.
        Called by the :meth:`__selectedOverlayChanged` and
        :meth:`__overlayNameChanged` methods.
        """

        overlay   = self.__currentOverlay
        display   = self.__currentDisplay
        infoFunc  = '_{}__get{}Info'.format(type(self)   .__name__,
                                            type(overlay).__name__)
        infoFunc  = getattr(self, infoFunc, None)

        # Overlay is none, or the overlay
        # type is not supported
        if infoFunc is None:
            self.__info.SetPage('', '')
            self.__info.Refresh()
            return

        info = infoFunc(overlay, display)

        self.__info.SetPage(self.__formatOverlayInfo(info), '')
        self.__info.Refresh()


    def __getImageInfo(self, overlay, display, title=None, metadata=True):
        """Creates and returns an :class:`OverlayInfo` object containing
        information about the given :class:`.Image` overlay.

        :arg overlay: A :class:`.Image` instance.
        :arg display: The :class:`.Display` instance assocated with the
                      ``Image``.
        """

        img     = overlay.nibImage
        hdr     = overlay.header
        isNifti = overlay.niftiVersion >= 1
        hasMeta = metadata and len(overlay.metaKeys()) > 0
        opts    = display.opts

        if isNifti: title = strings.labels[self, overlay]
        else:       title = strings.labels[self, 'Analyze']

        if overlay.dataSource is not None and \
           fslbids.isBIDSFile(overlay.dataSource):
            metaSect = strings.labels[self, overlay, 'bidsMeta']
        else:
            metaSect = strings.labels[self, overlay, 'jsonMeta']

        info = OverlayInfo('{} - {}'.format(display.name, title))

        generalSect = strings.labels[self,          'general']
        dimSect     = strings.labels[self, overlay, 'dimensions']
        xformSect   = strings.labels[self, overlay, 'transform']
        orientSect  = strings.labels[self, overlay, 'orient']

        info.addSection(generalSect)
        info.addSection(dimSect)
        info.addSection(xformSect)
        info.addSection(orientSect)

        if hasMeta:
            info.addSection(metaSect)

        displaySpace = strings.labels[self,
                                      overlay,
                                      'displaySpace',
                                      opts.transform]

        if opts.transform == 'reference':
            dsImg = self.displayCtx.displaySpace
            if isinstance(dsImg, fslimage.Nifti):
                dsDisplay    = self.displayCtx.getDisplay(dsImg)
                displaySpace = displaySpace.format(dsDisplay.name)
            else:
                log.warn('{} transform ({}) seems to be out '
                         'of date (display space: {})'.format(
                             overlay,
                             opts.transform,
                             self.displayCtx.displaySpace))

        dataType = strings.nifti.get(('datatype', int(hdr['datatype'])),
                                     'Unknown')

        info.addInfo(strings.labels[self, 'niftiVersion'],
                     strings.nifti['version.{}'.format(overlay.niftiVersion)],
                     section=generalSect)
        info.addInfo(strings.labels[self, 'dataSource'],
                     overlay.dataSource,
                     section=generalSect)
        info.addInfo(strings.nifti['datatype'],
                     dataType,
                     section=generalSect)
        info.addInfo(strings.nifti['descrip'],
                     overlay.strval('descrip'),
                     section=generalSect)

        if isNifti:
            intent = strings.nifti.get(
                ('intent_code', int(hdr['intent_code'])),
                'Unknown')

            info.addInfo(strings.nifti['intent_code'],
                         intent,
                         section=generalSect)
            info.addInfo(strings.nifti['intent_name'],
                         overlay.strval('intent_name'),
                         section=generalSect)

        info.addInfo(strings.nifti['aux_file'],
                     overlay.strval('aux_file'),
                     section=generalSect)

        info.addInfo(strings.labels[self, 'overlayType'],
                     strings.choices[display, 'overlayType'][
                         display.overlayType],
                     section=generalSect)
        info.addInfo(strings.labels[self, 'displaySpace'],
                     displaySpace,
                     section=generalSect)

        info.addInfo(strings.nifti['dimensions'],
                     '{}D'.format(len(overlay.shape)),
                     section=dimSect)

        for i in range(len(overlay.shape)):
            info.addInfo(strings.nifti['dim{}'.format(i + 1)],
                         str(overlay.shape[i]),
                         section=dimSect)

        # NIFTI images can specify different units.
        if isNifti:
            voxUnits  = overlay.xyzUnits
            timeUnits = overlay.timeUnits

        # Assume mm / seconds for ANALYZE images
        # We are using nifti xyzt_unit code values
        # here
        else:
            voxUnits, timeUnits = 2, 8

        # Convert the unit codes into labels
        voxUnits  = strings.nifti.get(('xyz_unit', voxUnits),  'INVALID CODE')
        timeUnits = strings.nifti.get(('t_unit',   timeUnits), 'INVALID CODE')

        for i in range(len(overlay.shape)):

            pixdim = hdr['pixdim'][i + 1]

            if   i  < 3: pixdim = '{:0.4g} {}'.format(pixdim, voxUnits)
            elif i == 3: pixdim = '{:0.4g} {}'.format(pixdim, timeUnits)

            info.addInfo(
                strings.nifti['pixdim{}'.format(i + 1)],
                pixdim,
                section=dimSect)

        bounds = affine.axisBounds(overlay.shape[:3],
                                   opts.getTransform('voxel', 'world'))
        lens   = bounds[1] - bounds[0]
        lens   = 'X={:0.0f} {} Y={:0.0f} {} Z={:0.0f} {}'.format(
            lens[0], voxUnits,
            lens[1], voxUnits,
            lens[2], voxUnits)

        info.addInfo(
            strings.labels[self, overlay, 'size'],
            lens,
            section=dimSect)

        # For NIFTI images, we show both
        # the sform and qform matrices,
        # in addition to the effective
        # transformation
        if isNifti:

            qformCode = int(hdr['qform_code'])
            sformCode = int(hdr['sform_code'])

            info.addInfo(strings.nifti['transform'],
                         self.__formatArray(overlay.voxToWorldMat),
                         section=xformSect)

            info.addInfo(strings.nifti['sform_code'],
                         strings.anatomy['Nifti', 'space', sformCode],
                         section=xformSect)
            info.addInfo(strings.nifti['qform_code'],
                         strings.anatomy['Nifti', 'space', qformCode],
                         section=xformSect)

            if sformCode != constants.NIFTI_XFORM_UNKNOWN:
                sform = img.get_sform()
                info.addInfo(strings.nifti['sform'],
                             self.__formatArray(sform),
                             section=xformSect)

            if qformCode != constants.NIFTI_XFORM_UNKNOWN:
                try:
                    qform = img.get_qform()
                except Exception as e:
                    log.warning('Could not read qform from {}: {}'.format(
                        overlay.name, str(e)))
                    qform = np.eye(4) * np.nan
                info.addInfo(strings.nifti['qform'],
                             self.__formatArray(qform),
                             section=xformSect)

        # For ANALYZE images, we show
        # the scale/offset matrix
        else:
            info.addInfo(strings.nifti['affine'],
                         self.__formatArray(hdr.get_best_affine()),
                         section=xformSect)

        if overlay.getXFormCode() == constants.NIFTI_XFORM_UNKNOWN:
            storageOrder = 'unknown'
        elif overlay.isNeurological(): storageOrder = 'neuro'
        else:                          storageOrder = 'radio'
        storageOrder = strings.nifti['storageOrder.{}'.format(storageOrder)]

        info.addInfo(strings.nifti['storageOrder'],
                     storageOrder,
                     section=orientSect)

        for i in range(3):
            xform  = opts.getTransform('voxel', 'world')
            orient = overlay.getOrientation(i, xform)
            orient = '{} - {}'.format(
                strings.anatomy['Nifti', 'lowlong',  orient],
                strings.anatomy['Nifti', 'highlong', orient])
            info.addInfo(strings.nifti['voxOrient.{}'.format(i)],
                         orient,
                         section=orientSect)

        for i in range(3):
            xform  = np.eye(4)
            orient = overlay.getOrientation(i, xform)
            orient = '{} - {}'.format(
                strings.anatomy['Nifti', 'lowlong',  orient],
                strings.anatomy['Nifti', 'highlong', orient])
            info.addInfo(strings.nifti['worldOrient.{}'.format(i)],
                         orient,
                         section=orientSect)

        if hasMeta:
            for k, v in overlay.metaItems():
                info.addInfo(k, str(v), metaSect)

        return info


    def __getFEATImageInfo(self, overlay, display):
        """Creates and returns an :class:`OverlayInfo` object containing
        information about the given :class:`.FEATImage` overlay.

        :arg overlay: A :class:`.FEATImage` instance.
        :arg display: The :class:`.Display` instance assocated with the
                      ``FEATImage``.
        """

        info = self.__getImageInfo(overlay, display)

        featInfo = [
            ('analysisName',   overlay.getAnalysisName()),
            ('analysisDir',    overlay.getFEATDir()),
            ('numPoints',      overlay.numPoints()),
            ('numEVs',         overlay.numEVs()),
            ('numContrasts',   overlay.numContrasts())]

        topLevel = overlay.getTopLevelAnalysisDir()
        report   = overlay.getReportFile()

        if topLevel is not None:
            featInfo.insert(2, ('partOfAnalysis', topLevel))

        if report is not None:
            report = '<a href="file://{}">{}</a>'.format(report, report)
            featInfo.insert(2, ('report', report))

        secName = strings.labels[self, overlay, 'featInfo']
        info.addSection(secName)

        for k, v in featInfo:
            info.addInfo(strings.feat[k], v, section=secName)

        return info


    def __getMelodicImageInfo(self, overlay, display):
        """Creates and returns an :class:`OverlayInfo` object containing
        information about the given :class:`.MelodicImage` overlay.

        :arg overlay: A :class:`.MelodicImage` instance.
        :arg display: The :class:`.Display` instance assocated with the
                      ``MelodicImage``.
        """

        info = self.__getImageInfo(overlay, display)

        melInfo = [
            ('dataFile',       overlay.getDataFile()),
            ('analysisDir',    overlay.getMelodicDir()),
            ('tr',             overlay.tr),
            ('numComponents',  overlay.numComponents())]

        topLevel = overlay.getTopLevelAnalysisDir()
        report   = overlay.getReportFile()

        if topLevel is not None:
            melInfo.insert(2, ('partOfAnalysis', topLevel))

        if report is not None:
            report = '<a href="file://{}">{}</a>'.format(report, report)
            melInfo.insert(2, ('report', report))

        secName = strings.labels[self, overlay, 'melodicInfo']
        info.addSection(secName)

        for k, v in melInfo:
            info.addInfo(strings.melodic[k], v, section=secName)

        return info


    def __getMeshInfo(self, overlay, display):
        """Creates and returns an :class:`OverlayInfo` object containing
        information about the given :class:`.Mesh` overlay.

        :arg overlay: A :class:`.Mesh` instance.
        :arg display: The :class:`.Display` instance assocated with the
                      ``Mesh``.
        """

        opts   = display.opts
        refImg = opts.refImage

        modelInfo = [
            ('numVertices',  overlay.vertices.shape[0]),
            ('numTriangles', overlay.indices .shape[0]),
        ]

        if refImg is None:
            modelInfo.append(
                ('displaySpace', strings.labels[
                    self, overlay, 'coordSpace', 'display']))

            mesh2worldXform = np.eye(4)

        else:

            refOpts         = self.displayCtx.getOpts(refImg)
            dsImg           = self.displayCtx.displaySpace
            displaySpace    = strings.labels[
                self, refImg, 'displaySpace', refOpts.transform]
            coordSpace      = strings.labels[
                self, overlay,
                'coordSpace', opts.coordSpace].format(refImg.name)
            mesh2worldXform = affine.concat(
                refOpts.getTransform('display', 'world'),
                opts.getTransform('mesh', 'display'))

            if refOpts.transform == 'reference':
                dsDisplay    = self.displayCtx.getDisplay(dsImg)
                displaySpace = displaySpace.format(dsDisplay.name)

            modelInfo.append(('refImage',     refImg.dataSource))
            modelInfo.append(('coordSpace',   coordSpace))
            modelInfo.append(('displaySpace', displaySpace))

        bounds = affine.transform(overlay.bounds, mesh2worldXform)
        lens   = bounds[1] - bounds[0]
        lens   = 'X={:0.0f} mm Y={:0.0f} mm Z={:0.0f} mm'.format(*lens)

        modelInfo.append(('size', lens))

        info = OverlayInfo('{} - {}'.format(
            display.name,
            strings.labels[self, overlay]))

        info.addInfo(strings.labels[self, 'dataSource'], overlay.dataSource)

        for name, value in modelInfo:
            info.addInfo(strings.labels[self, overlay, name], value)

        return info


    def __getVTKMeshInfo(self, overlay, display):
        """Creates and returns an :class:`OverlayInfo` object containing
        information about the given :class:`.VTKMesh` overlay.

        :arg overlay: A :class:`.VTKMesh` instance.
        :arg display: The :class:`.Display` instance assocated with the
                      ``VTKMesh``.
        """

        info       = self.__getMeshInfo(overlay, display)
        info.title = strings.labels[self, overlay]

        return info


    def __getGiftiMeshInfo(self, overlay, display):
        """Creates and returns an :class:`OverlayInfo` object containing
        information about the given :class:`.GiftiMesh` overlay.

        :arg overlay: A :class:`.GiftiMesh` instance.
        :arg display: The :class:`.Display` instance assocated with the
                      ``GiftiMesh``.
        """

        info       = self.__getMeshInfo(overlay, display)
        info.title = strings.labels[self, overlay]

        return info


    def __getFreesurferMeshInfo(self, overlay, display):
        """Creates and returns an :class:`OverlayInfo` object containing
        information about the given :class:`.FreesurferMesh` overlay.

        :arg overlay: A :class:`.FreesurferMesh` instance.
        :arg display: The :class:`.Display` instance assocated with the
                      ``FreesurferMesh``.
        """

        info       = self.__getMeshInfo(overlay, display)
        info.title = strings.labels[self, overlay]

        return info


    def __getDTIFitTensorInfo(self, overlay, display):
        """Creates and returns an :class:`OverlayInfo` object containing
        information about the given :class:`.DTIFitTensor` overlay.

        :arg overlay: A :class:`.DTIFitTensor` instance.
        :arg display: The :class:`.Display` instance assocated with the
                      ``DTIFitTensor``.
        """
        info = self.__getImageInfo(overlay.L1(), display)

        generalSect = strings.labels[self, 'general']
        dataSource  = strings.labels[self, 'dataSource']

        info.title = strings.labels[self, overlay]
        info.sections[generalSect][dataSource] = overlay.dataSource

        tensorInfo = [
            ('v1', overlay.V1().dataSource),
            ('v2', overlay.V2().dataSource),
            ('v3', overlay.V3().dataSource),
            ('l1', overlay.L1().dataSource),
            ('l2', overlay.L2().dataSource),
            ('l3', overlay.L3().dataSource),
        ]

        section = strings.labels[self, overlay, 'tensorInfo']

        info.addSection(section)

        for name, val in tensorInfo:
            info.addInfo(strings.tensor[name], val, section)

        return info


    def __getDicomImageInfo(self, overlay, display):
        """Creates and returns an :class:`OverlayInfo` object containing
        information about the given :class:`.DicomImage` overlay.

        :arg overlay: A :class:`.DicomImage` instance.
        :arg display: The :class:`.Display` instance assocated with the
                      ``DicomImage``.
        """

        info      = self.__getImageInfo(overlay, display, metadata=False)
        dicomInfo = strings.labels[self, overlay, 'dicomMeta']

        info.addInfo(strings.labels[self, overlay, 'dicomDir'],
                     overlay.dicomDir)

        info.addSection(dicomInfo)

        for k, v in overlay.metaItems():
            info.addInfo(k, str(v), dicomInfo)

        return info


    def __getMGHImageInfo(self, overlay, display):
        """Creates and returns an :class:`OverlayInfo` object containing
        information about the given :class:`.MGHImage` overlay.

        :arg overlay: A :class:`.MGHImage` instance.
        :arg display: The :class:`.Display` instance assocated with the
                      ``DicomImage``.
        """

        info     = self.__getImageInfo(overlay, display)
        filename = overlay.mghImageFile

        if filename is not None:
            info.addInfo(strings.labels[self, overlay, 'filename'], filename)

        return info


    def __formatArray(self, array):
        """Creates and returns a string containing a HTML table which
        formats the data in the given ``numpy.array``.
        """

        lines = []

        lines.append('<table border="0" style="font-size: small;">')

        for rowi in range(array.shape[0]):

            lines.append('<tr>')

            for coli in range(array.shape[1]):
                lines.append('<td>{:0.4g}</td>'.format(array[rowi, coli]))
            lines.append('</tr>')

        lines.append('</table>')

        return ''.join(lines)


    def __formatOverlayInfo(self, info):
        """Creates and returns a string containing some HTML which formats
        the information in the given ``OverlayInfo`` instance.
        """
        lines = []

        lines.append('<html>')
        lines.append('<body style="font-family: '
                     'sans-serif; font-size: small;">')
        lines.append('<h2>{}</h2>'.format(info.title))

        sections = []

        if len(info.info) > 0:
            sections.append((None, list(info.info.items())))

        for secName, secInf in info.sections.items():
            sections.append((secName, list(secInf.items())))

        for i, (secName, secInf) in enumerate(sections):

            lines.append('<div style="float:left; margin: 5px; '
                         'background-color: #f0f0f0;">')

            if secName is not None:
                lines.append('<h3>{}</h3>'.format(secName))

            lines.append('<table border="0" style="font-size: small;">')

            for i, (infName, infData) in enumerate(secInf):

                if i % 2: bgColour = '#f0f0f0'
                else:     bgColour = '#cdcdff'

                lines.append('<tr bgcolor="{}">'
                             '<td><b>{}</b></td>'
                             '<td>{}</td></tr>'.format(
                                 bgColour,
                                 infName,
                                 infData))

            lines.append('</table>')
            lines.append('</div>')


        lines.append('</body></html>')

        return '\n'.join(lines)


class OverlayInfo(object):
    """A little class which encapsulates human-readable information about
    one overlay. ``OverlayInfo`` objects are created and returned by the
    ``OverlayInfoPanel.__get*Info`` methods.

    The information stored in an ``OverlayInfo`` instance is organised into
    *sections*. Within each section, information is organised into key-value
    pairs. The order in which both ``OverlayInfo`` sections, and information,
    is ultimately output, is the order in which the sections/information are
    added, via the :meth:`addSection` and :meth:`addInfo` methods.
    """

    def __init__(self, title):
        """Create an ``OverlayInfo`` instance.

        :arg title: The ``OverlaytInfo`` title.
        """

        self.title    = title
        self.info     = collections.OrderedDict()
        self.sections = collections.OrderedDict()


    def addSection(self, section):
        """Add a section to this ``OverlayInfo`` instance.

        :arg section: The section name.
        """
        self.sections[section] = collections.OrderedDict()


    def addInfo(self, name, info, section=None):
        """Add some information to this ``OverlayInfo`` instance.

        :arg name:    The information name.
        :arg info:    The information value.
        :arg section: Section to place the information in.
        """
        if section is None: self.info[             name] = info
        else:               self.sections[section][name] = info



#!/usr/bin/env python
#
# addmaskdataseries.py - The AddMaskDataSeriesAction class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`AddMaskDataSeriesAction` class, an action
used by the :class:`.TimeSeriesPanel`.
"""


import          wx
import numpy as np

import fsl.data.image              as fslimage

import fsleyes.strings             as strings
import fsleyes.plotting.dataseries as dataseries
from . import                         base


class AddMaskDataSeriesAction(base.Action):
    """The ``AddMaskDataSeriesAction`` class is used by the
    :class:`.TimeSeriesPanel`.

    It prompts the user to select a mask image for the currently selected
    overlay (assumed to be a 4D time series :class:`.Image`), then extracts
    the mean time series for the non-zero voxels within the mask, and adds
    them as a :class:`.DataSeries` to the ``TimeSeriesPanel``.
    """

    def __init__(self, overlayList, displayCtx, plotPanel):
        """Create an ``AddMaskDataSeriesAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg plotPanel:   The :class:`.TimeSeriesPanel`.
        """

        base.Action.__init__(
            self, overlayList, displayCtx, self.__addMaskDataSeries)

        self.__plotPanel   = plotPanel
        self.__name        = '{}_{}'.format(type(self).__name__, id(self))
        self.__maskOptions = []

        overlayList.addListener('overlays',
                                self.__name,
                                self.__overlayListChanged)
        displayCtx .addListener('selectedOverlay',
                                self.__name,
                                self.__overlayListChanged)

        self.__overlayListChanged()


    def destroy(self):
        """Must be called when this ``AddMaskDataSeriesAction`` is no
        longer in use.
        """
        self.overlayList.removeListener('overlays',        self.__name)
        self.displayCtx .removeListener('selectedOverlay', self.__name)
        self.__plotPanel   = None
        self.__maskOptions = None
        base.Action.destroy(self)


    def __overlayListChanged(self, *a):
        """Called when the :class:`.OverlayList` changes. Updates the
        :attr:`.Action.enabled` flag based on the currently selected
        overlay, and the contents of the overlay list.
        """

        overlay = self.displayCtx.getSelectedOverlay()

        if (len(self.overlayList) == 0 or
           (not isinstance(overlay, fslimage.Image))):
            self.enabled = False
            return

        self.__maskOptions = [o for o in self.overlayList if
                              isinstance(o, fslimage.Image) and
                              o is not overlay              and
                              o.sameSpace(overlay)]

        self.enabled = (overlay.ndim > 3 and len(self.__maskOptions) > 0)


    def __addMaskDataSeries(self):
        """Run the ``AddMaskDataSeriesAction``. Prompt the user to select
        a mask, using a :class:`MaskDialog`, then calculates the mean time
        series in that mask, then adds that time series to the
        :class:`.TimeSeriesPanel` that owns this action instance.
        """

        overlay = self.displayCtx.getSelectedOverlay()
        opts    = self.displayCtx.getOpts(overlay)
        options = self.__maskOptions

        frame   = wx.GetApp().GetTopWindow()
        msg     = strings.messages[self, 'selectMask'].format(overlay.name)
        cbmsg   = strings.messages[self, 'weighted']
        title   = strings.titles[  self, 'selectMask'].format(overlay.name)

        dlg = MaskDialog(
            frame,
            [o.name for o in options],
            title=title,
            message=msg,
            checkboxMessage=cbmsg)

        if dlg.ShowModal() != wx.ID_OK:
            return

        maskimg   = options[dlg.GetChoice()]
        weight    = dlg.GetCheckBox()
        ds        = dataseries.DataSeries(overlay,
                                          self.overlayList,
                                          self.displayCtx,
                                          self.__plotPanel)

        data     = overlay.data[opts.index(atVolume=False)]
        mask     = maskimg.data
        maskmask = mask > 0
        ydata    = data[maskmask]

        # Weighted mean
        if weight:
            maskvals = mask[maskmask]
            ydata    = (maskvals * ydata.T).T

        ydata = ydata.mean(axis=0)
        xdata = np.arange(len(ydata))

        ds.colour    = self.__plotPanel.getOverlayPlotColour(overlay)
        ds.lineStyle = self.__plotPanel.getOverlayPlotStyle(overlay)
        ds.lineWidth = 2
        ds.alpha     = 1
        ds.label     = '{} [mask: {}]'.format(overlay.name, maskimg.name)

        # We have to run the data through
        # prepareDataSeries to e.g. scale
        # the x axis by pixdims, and apply
        # other plot settings
        ds.setData(xdata, ydata)
        ds.setData(*self.__plotPanel.prepareDataSeries(ds))

        self.__plotPanel.dataSeries.append(ds)



class MaskDialog(wx.Dialog):
    """A dialog which displays some options to the user:

     - A ``Choice`` widget containing a list of mask images
     - A checkbox allowing the user to select whether to calculate
       the weighted mean time series, weighted by the mask values,
       or calculate the unweighted mean.

    The selections are available via the :meth:`GetMask` and
    :meth:`GetWeighted` methods
    """

    def __init__(self,
                 parent,
                 choices,
                 title=None,
                 message=None,
                 checkbox=True,
                 checkboxMessage=None):
        """Create a ``ChoiceDialog``.

        :arg parent:          ``wx`` parent object.
        :arg choices:         List of strings, the choices to present to the
                              user.
        :arg title:           Dialog title
        :arg message:         Message to show above choice widget.
        :arg checkbox:        Show a checkbox
        :arg checkboxMessage: Message to show alongside checkbox widget.
        """

        if title           is None: title           = ''
        if message         is None: message         = ''
        if checkboxMessage is None: checkboxMessage = ''

        wx.Dialog.__init__(self,
                           parent,
                           title=title,
                           style=wx.DEFAULT_DIALOG_STYLE)

        self.__message      = wx.StaticText(self, label=message)
        self.__choice       = wx.Choice(self,     choices=choices)

        if checkbox: self.__checkbox = wx.CheckBox(self, label=checkboxMessage)
        else:        self.__checkbox = None

        self.__okButton     = wx.Button(self, label='Ok',     id=wx.ID_OK)
        self.__cancelButton = wx.Button(self, label='Cancel', id=wx.ID_CANCEL)

        self.__okButton    .Bind(wx.EVT_BUTTON, self.__onOkButton)
        self.__cancelButton.Bind(wx.EVT_BUTTON, self.__onCancelButton)

        self.__okButton.SetDefault()

        self.__mainSizer   = wx.BoxSizer(wx.VERTICAL)
        self.__buttonSizer = wx.BoxSizer(wx.HORIZONTAL)

        self.__buttonSizer.Add((1, 1), flag=wx.EXPAND, proportion=1)
        self.__buttonSizer.Add(self.__okButton)
        self.__buttonSizer.Add((5, 1), flag=wx.EXPAND)
        self.__buttonSizer.Add(self.__cancelButton)

        self.__mainSizer.Add(self.__message,
                             flag=wx.EXPAND | wx.ALL,
                             proportion=1,
                             border=20)
        self.__mainSizer.Add(self.__choice,
                             flag=wx.EXPAND | wx.LEFT | wx.RIGHT,
                             border=20)

        if checkbox:
            self.__mainSizer.Add(self.__checkbox,
                                 flag=wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP,
                                 border=20)

        self.__mainSizer.Add(self.__buttonSizer,
                             flag=wx.EXPAND | wx.ALL,
                             border=20)

        self.SetSizer(self.__mainSizer)
        self.Layout()
        self.Fit()
        self.CentreOnParent()


    @property
    def okButton(self):
        """Returns the OK button. """
        return self.__okButton


    @property
    def cancelButton(self):
        """Returns the cancel button. """
        return self.__cancelButton


    @property
    def checkbox(self):
        """Returns the checkbox. """
        return self.__checkbox


    @property
    def choice(self):
        """Returns the choice widget. """
        return self.__choice


    def GetChoice(self):
        """Returns the index of the currently selected choice."""
        return self.__choice.GetSelection()


    def GetCheckBox(self):
        """Returns the index of the currently selected choice."""
        if self.__checkbox is None:
            raise RuntimeError('This dialog does not have a checkbox')

        return self.__checkbox.GetValue()


    def __onOkButton(self, ev):
        """Called when the ok button is pushed. """
        self.EndModal(wx.ID_OK)


    def __onCancelButton(self, ev):
        """Called when the cancel button is pushed. """
        self.EndModal(wx.ID_CANCEL)
